# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import numpy as np
from functools import reduce
from math import pi
from collections.abc import Iterable, Sequence

from netket.utils import struct, deprecated_new_name, deprecated
from netket.utils.types import Array
from netket.utils.float import prune_zeros
from netket.utils.dispatch import dispatch
from netket.utils.group import (
    Element,
    Identity,
    PointGroup,
    Permutation,
    PermutationGroup,
)

from .lattice import Lattice


class Translation(Permutation):
    r"""
    Custom subclass of `Permutation` that represents a lattice permutation.
    Stores translation lattice vector and generates a sensible name from it.

    The product of two `Translation`s carries the appropriate displacement vector.
    """

    def __init__(
        self,
        permutation: Array | None = None,
        *,  # maybe change something
        displacement: Array,
        permutation_array: Array | None = None,
        inverse_permutation_array: Array | None = None,
    ):
        r"""
        Creates a `Translation` from either the array of images
        `permutation_array` or preimages `inverse_permutation_array`,
        and a displacement vector.

        Exactly one argument among `permutation_array` and
        `inverse_permutation_array` (and the deprecated argument `permutation`)
        must be specified.

        The deprecated argument `permutation` should be substituted for
        `inverse_permutation_array`.

        Note that the left action of a permutation on an array `a` is
        `a[inverse_permutation_array]`.

        Args:
            permutation: (deprecated) 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.
            displacement: displacement vector is units of lattice basis vectors
            permutation_array: 1D array listing
                :math:`g(x)` for all :math:`0\le x \le N-1`.
            inverse_permutation_array: 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.

        Returns:
            A `Translation` object that encodes the specified translation.
        """
        super().__init__(
            permutation_array=permutation_array,
            inverse_permutation_array=inverse_permutation_array,
            permutation=permutation,
        )
        self._vector = np.asarray(displacement)

    @property
    def _name(self):
        return f"Translation({self._vector.tolist()})"


@dispatch
def product(p: Translation, q: Translation):
    inverse_permutation_array = q.inverse_permutation_array[p.inverse_permutation_array]
    return Translation(
        inverse_permutation_array=inverse_permutation_array,
        displacement=p._vector + q._vector,
    )


def _ensure_iterable(x):
    """Extracts iterables given in varargs"""
    if isinstance(x[0], Iterable):
        if len(x) > 1:
            raise TypeError("Either Iterable or variable argument list expected")
        return x[0]
    else:
        return x


# This function doesn't seem to be tested
def _translations_along_axis(lattice: Lattice, axis: int) -> PermutationGroup:
    """
    The group of valid translations along an axis as a `PermutationGroup`
    acting on the sites of `lattice.`
    """
    if lattice._pbc[axis]:
        trans_list = [Identity()]
        # note that we need the preimages in the permutation
        trans_perm = lattice.id_from_position(
            lattice.positions - lattice.basis_vectors[axis]
        )
        vector = np.zeros(lattice.ndim, dtype=int)
        vector[axis] = 1
        trans_by_one = Translation(
            inverse_permutation_array=trans_perm, displacement=vector
        )

        for _ in range(1, lattice.extent[axis]):
            trans_list.append(trans_list[-1] @ trans_by_one)

        return PermutationGroup(trans_list, degree=lattice.n_nodes)
    else:
        return PermutationGroup([Identity()], degree=lattice.n_nodes)


# This function doesn't seem to be tested
def _pg_to_permutation(lattice: Lattice, point_group: PointGroup) -> PermutationGroup:
    """
    The permutation action of `point_group` on the sites of `lattice`.
    """
    perms: list[Element] = []
    for p in point_group:
        if isinstance(p, Identity):
            perms.append(Identity())
        else:
            # note that we need the preimages in the permutation
            perm = lattice.id_from_position(p.preimage(lattice.positions))
            perms.append(Permutation(inverse_permutation_array=perm, name=str(p)))
    return PermutationGroup(perms, degree=lattice.n_nodes)


@struct.dataclass
class TranslationGroup(PermutationGroup):
    """
    Class to handle translation symmetries of a `Lattice`. Corresponds to a representation of the translation group
    on the given lattice as a permutation group of `N_sites` variables.

    Can be used as a `PermutationGroup` representing the translations,
    but the product table is computed much more efficiently than a generic
    `PermutationGroup`.
    """

    lattice: Lattice
    """The lattice whose translation group is represented."""
    axes: tuple[int]
    """Axes translations along which are represented by the group."""

    def __repr__(self):
        return type(self).__name__ + f"(lattice:\n{self.lattice}\naxes:{self.axes})"

    def __pre_init__(
        self, lattice: Lattice, axes: int | tuple[int] | None = None
    ) -> tuple[tuple, dict]:
        if axes is None:
            axes = tuple(range(lattice.ndim))
        elif isinstance(axes, int):
            axes = [axes]
        else:
            assert all(x < lattice.ndim for x in axes)
            assert len(set(axes)) == len(axes)

        # compute translation group by axis and overall
        translation_by_axis = [_translations_along_axis(lattice, i) for i in axes]
        translation_group = reduce(PermutationGroup.__matmul__, translation_by_axis)

        return (), dict(
            lattice=lattice,
            axes=tuple(axes),
            elems=translation_group.elems,
            degree=lattice.n_nodes,
        )

    def __hash__(self):
        return super().__hash__()

    @property
    def group_shape(self) -> Array:
        shape = [l if p else 1 for (l, p) in zip(self.lattice.extent, self.lattice.pbc)]
        return np.asarray(shape)

    @struct.property_cached
    def inverse(self) -> Array:
        ix = np.ix_(*[-np.arange(x) % x for x in self.group_shape])
        inv = np.arange(len(self)).reshape(self.group_shape)[ix]
        return np.ravel(inv)

    @struct.property_cached
    def product_table(self) -> Array:
        # product table of each 1d component
        ix = [(np.arange(x) - np.arange(x)[:, None]) % x for x in self.group_shape]
        ix = np.ix_(*[np.ravel(x) for x in ix])
        # product table with n_axes axes
        # each axis stands for row and column direction along one axis
        pt = np.arange(len(self)).reshape(self.group_shape)[ix]
        shape = [x for x in self.group_shape for _ in range(2)]
        # separate row and column directions
        pt = pt.reshape(shape)
        # bring all row and all column directions together
        pt = pt.transpose(list(range(0, len(shape), 2)) + list(range(1, len(shape), 2)))

        return pt.reshape(len(self), len(self))


@struct.dataclass
class SpaceGroup(PermutationGroup):
    """
    Class to handle the space group symmetries of `Lattice`.

    Can be used as a `PermutationGroup` representing the action of a
    space group on a `Lattice`. The space group is generated as the
    semidirect product of the translation group of the `Lattice` and
    a geometrical point group given as a constructor argument.

    Also generates `PermutationGroup` representations of
    * the supplied point group,
    * its rotational subgroup (i.e. point group symmetries with determinant +1)
    * the translation group of the `Lattice`

    Also generates space group irreps for symmetrising wave functions.
    """

    lattice: Lattice
    """The lattice underlying the space group."""
    _point_group: PointGroup
    """The geometric point group underlying the space group."""
    point_group: PermutationGroup
    """The point group as a `PermutationGroup` acting on the sites of `self.lattice`.

    Group elements are listed in the order they appear in `self._point_group`.
    Computed from `_point_group` upon construction, must not be changed after."""
    full_translation_group: PermutationGroup

    def __pre_init__(
        self, lattice: Lattice, point_group: PointGroup
    ) -> tuple[tuple, dict]:
        """
        Constructs the Space Group Builder used to concretize a point group
        which knows nothing about how many sites there are in a lattice,
        into a Permutation Group which can be used to perform calculations.

        From the point of view of group theory, you can think of this as
        taking the point group and returning a representation on the computational
        basis defined by the lattice.

        Args:
            lattice: The lattice for which to represent the point group as
                a permutation group
            point_group: The point group to be represented
        """

        if not isinstance(lattice, Lattice):
            raise TypeError(
                "Expected an instance of `Lattice` as argument `lattice`, "
                f"got {type(lattice)}."
            )

        if not isinstance(point_group, PointGroup):
            msg = (
                "Expected an instance of `PointGroup` as argument `point_group`, "
                f"got type {type(point_group)}."
            )
            if isinstance(point_group, Element):
                msg += (
                    "\n\n`point_group` appears to be a single symmetry. "
                    "It should be wrapped into a `PointGroup` object."
                )
            raise TypeError(msg)

        # compute point group permutations
        point_group = point_group.replace(unit_cell=lattice.basis_vectors)
        pg_as_perm = _pg_to_permutation(lattice, point_group)

        # compute translation group by axis and overall
        translation_group = lattice.full_translation_group

        # compute space group elements
        space_group = translation_group @ pg_as_perm
        elems = space_group.elems

        return (), dict(
            lattice=lattice,
            _point_group=point_group,
            point_group=pg_as_perm,
            full_translation_group=translation_group,
            elems=elems,
            degree=lattice.n_nodes,
        )

    def __hash__(self):
        return super().__hash__()

    def __repr__(self):
        return (
            type(self).__name__
            + f"(lattice:\n{self.lattice}\npoint_group:\n{self._point_group})"
        )

    @property
    @deprecated_new_name("_point_group", reason="Consistency")
    def point_group_(self) -> PointGroup:
        """
        Deprecated: Returns the internally stored point group as a point group,
        instead of the one stored as a permutation group.
        """

        return self._point_group

    # TODO describe ordering of group elements here and later in docstring

    @struct.property_cached
    def rotation_group(self) -> PermutationGroup:
        """The group of rotations (i.e. point group symmetries with determinant +1)
        as a `PermutationGroup` acting on the sites of `self.lattice`.

        Group elements are listed in the order they appear in `self._point_group`."""
        return _pg_to_permutation(self.lattice, self._point_group.rotation_group())

    def translation_group(
        self, axes: int | Sequence[int] | None = None
    ) -> TranslationGroup:
        """
        The group of valid translations of `self.lattice` as a `PermutationGroup`
        acting on the sites of the same.
        """
        if axes is None:
            return self.full_translation_group
        else:
            return TranslationGroup(self.lattice, axes=axes)

    @property
    @deprecated(
        reason="This `SpaceGroup` object can be used directly as a permutation group"
    )
    def space_group(self) -> "SpaceGroup":
        """
        The space group generated by `self.point_group` and `self.translation_group`.
        """
        return self

    @struct.property_cached
    def product_table(self) -> Array:
        # compute first n_PG rows of product table like in PermutationGroup
        perms = self.to_array()
        inverse = perms[self.inverse].squeeze()
        n_symm = len(perms)
        n_PG = len(self._point_group)
        n_TG = len(self.full_translation_group)
        lookup = np.unique(np.column_stack((perms, np.arange(len(self)))), axis=0)

        PG_rows = np.zeros([n_PG, n_symm], dtype=int)
        for i, g_inv in enumerate(inverse[:n_PG]):
            row_perms = perms[:, g_inv]
            row_perms = np.unique(
                np.column_stack((row_perms, np.arange(len(self)))), axis=0
            )
            # row_perms should be a permutation of perms, so identical after sorting
            if np.any(row_perms[:, :-1] != lookup[:, :-1]):
                raise RuntimeError(
                    "PermutationGroup is not closed under multiplication"
                )
            # match elements in row_perms to group indices
            PG_rows[i, row_perms[:, -1]] = lookup[:, -1]

        # PG_rows contains pg^-1 th ph - split three terms into three dimensions
        PG_rows = PG_rows.reshape(n_PG, n_TG, n_PG)
        # the full product table is of the form pg^-1 tg^-1 th ph
        # the middle two terms are the product table of the TG
        product_table = PG_rows[:, self.full_translation_group.product_table, :]
        # reshuffle into output shape
        product_table = product_table.transpose(1, 0, 2, 3)

        return product_table.reshape(n_symm, n_symm)

    def _little_group_index(self, k: Array) -> Array:
        """
        Returns the indices of the elements of the little group corresponding to
        wave vector `k`.
        """
        # calculate k' = p(k) for all p in the point group
        big_star = np.tensordot(self._point_group.matrices(), k, axes=1)
        big_star = self.lattice.to_reciprocal_lattice(big_star) % self.lattice.extent
        # should test for pbc before taking the modulus, but the only valid wave
        # vector for non-pbc axes is 0 and 0 % anything == 0

        # assumes point_group_[0] is the identity
        is_in_little_group = np.all(big_star == big_star[0], axis=1)
        return np.arange(len(self._point_group))[is_in_little_group]

    def little_group(self, *k: Array) -> PointGroup:
        """
        Returns the little co-group corresponding to wave vector *k*.
        This is the subgroup of the point group that leaves *k* invariant.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            the little co-group as a `PointGroup`
        """
        k = _ensure_iterable(k)
        return PointGroup(
            [self._point_group[i] for i in self._little_group_index(k)],
            ndim=self._point_group.ndim,
            unit_cell=self.lattice.basis_vectors,
        )

    def _little_group_irreps(self, k: Array, divide: bool = False) -> Array:
        """
        Returns the character table of the little group embedded in the full point
        group. Symmetries outside the little group get 0.
        If `divide` is `True`, the result gets divided by the size of the little group.
        This is convenient when calculating space group irreps.
        """
        idx = self._little_group_index(k)
        CT = self.little_group(k).character_table()
        CT_full = np.zeros((CT.shape[0], len(self._point_group)), dtype=CT.dtype)
        CT_full[:, idx] = CT
        return CT_full / idx.size if divide else CT_full

    def space_group_irreps(self, *k: Array) -> Array:
        """
        Returns the portion of the character table of the full space group corresponding
        to the star of the wave vector *k*.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            An array `CT` listing the characters for a number of irreps of the
            space group.
            `CT[i]` for each `i` gives a distinct irrep, each corresponding to
            `self.little_group(k).character_table[i].
            `CT[i,j]` gives the character of `self.space_group[j]` in the same.
        """
        k = _ensure_iterable(k)
        # Wave vectors
        big_star_Cart = np.tensordot(self._point_group.matrices(), k, axes=1)
        big_star = self.lattice.to_reciprocal_lattice(big_star_Cart) * (
            2 * pi / self.lattice.extent
        )
        # Little-group-irrep factors
        # Conjugacy_table[g,p] lists p^{-1}gp, so point_group_factors[i,:,p]
        #     of irrep #i for the little group of p(k) is the equivalent
        # Phase factor for non-symmorphic symmetries is exp(-i w_g . p(k))
        point_group_factors = self._little_group_irreps(k, divide=True)[
            :, self._point_group.conjugacy_table
        ] * np.exp(
            -1j
            * np.tensordot(
                self._point_group.translations(), big_star_Cart, axes=(-1, -1)
            )
        )
        # Translational factors
        trans_factors = []
        for axis in range(self.lattice.ndim):
            n_trans = self.lattice.extent[axis] if self.lattice.pbc[axis] else 1
            factors = np.exp(-1j * np.outer(np.arange(n_trans), big_star[:, axis]))
            shape = (
                [1] * axis
                + [n_trans]
                + [1] * (self.lattice.ndim - 1 - axis)
                + [len(self._point_group)]
            )
            trans_factors.append(factors.reshape(shape))
        trans_factors = reduce(np.multiply, trans_factors).reshape(
            -1, len(self._point_group)
        )

        # Multiply the factors together and sum over the "p" PGSymmetry axis
        # Translations are more major than point group operations
        result = np.einsum(
            "igp, tp -> itg", point_group_factors, trans_factors
        ).reshape(point_group_factors.shape[0], -1)
        return prune_zeros(result)

    def one_arm_irreps(self, *k: Array) -> Array:
        """
        Returns the portion of the character table of the full space group corresponding
        to the star of the wave vector *k*, projected onto *k* itself.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            An array `CT` listing the projected characters for a number of irreps of
            the space group.
            `CT[i]` for each `i` gives a distinct irrep, each corresponding to
            `self.little_group(k).character_table[i].
            `CT[i,j]` gives the character of `self.space_group[j]` in the same.
        """
        # Convert k to reciprocal lattice vectors
        k = _ensure_iterable(k)
        # Little-group irrep factors
        # Phase factor for non-symmorphic symmetries is exp(-i w_g . p(k))
        point_group_factors = self._little_group_irreps(k) * np.exp(
            -1j * (self._point_group.translations() @ k)
        )
        # Translational factors
        trans_factors = []
        for axis in range(self.lattice.ndim):
            n_trans = self.lattice.extent[axis] if self.lattice.pbc[axis] else 1
            factors = np.exp(-1j * k[axis] * np.arange(n_trans))
            shape = [1] * axis + [n_trans] + [1] * (self.lattice.ndim - 1 - axis)
            trans_factors.append(factors.reshape(shape))
        trans_factors = reduce(np.multiply, trans_factors).ravel()

        # Multiply the factors together
        # Translations are more major than point group operations
        result = np.einsum("ig, t -> itg", point_group_factors, trans_factors).reshape(
            point_group_factors.shape[0], -1
        )
        return prune_zeros(result)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from netket.utils.moduletools import export, hide_unexported

from .axial import cuboid, cuboid_rotations  # noqa: F401
from .axial import inversion_group as _inv_group
from .axial import reflection_group as _refl_group
from .axial import rotation as _rotation
from ._point_group import PointGroup
from ._semigroup import Identity


hide_unexported(__name__)

__all__ = [
    "tetrahedral_rotations",
    "tetrahedral",
    "pyritohedral",
    "octahedral_rotations",
    "cubic_rotations",
    "octahedral",
    "cubic",
    "diamond",
    "pyrochlore",
    "Fd3m",
]


@export
def T() -> PointGroup:
    """
    Rotational symmetries of a tetrahedron with vertices
    (1,1,1), (1,-1,-1), (-1,1,-1), (-1,-1,1).
    """
    return PointGroup(
        [
            Identity(),
            _rotation(120, [1, 1, 1]),
            _rotation(120, [1, -1, -1]),
            _rotation(120, [-1, 1, -1]),
            _rotation(120, [-1, -1, 1]),
            _rotation(-120, [1, 1, 1]),
            _rotation(-120, [1, -1, -1]),
            _rotation(-120, [-1, 1, -1]),
            _rotation(-120, [-1, -1, 1]),
            _rotation(180, [0, 0, 1]),
            _rotation(180, [0, 1, 0]),
            _rotation(180, [1, 0, 0]),
        ],
        ndim=3,
    )


tetrahedral_rotations = T


@export
def Td() -> PointGroup:
    r"""
    Symmetry group of a tetrahedron with vertices
    (1,1,1), (1,-1,-1), (-1,1,-1), (-1,-1,1).
    """
    return _refl_group([1, 1, 0]) @ T()


tetrahedral = Td


@export
def Th() -> PointGroup:
    """Pyritohedral symmetry group generated by T and inversion."""
    return _inv_group() @ T()


pyritohedral = Th


@export
def O() -> PointGroup:  # noqa: E741, E743
    """Rotational symmetries of a cube/octahedron aligned with the Cartesian axes."""
    # NB the first factor isn't an actual point group but this is fine
    # we only use it to generate a coset of T in O
    return PointGroup([Identity(), _rotation(90, [0, 0, 1])], ndim=3) @ T()


octahedral_rotations = cubic_rotations = O


@export
def Oh() -> PointGroup:
    """Symmetry group of a cube/octahedron aligned with the Cartesian axes."""
    return _inv_group() @ O()


octahedral = cubic = Oh


@export
def Fd3m() -> PointGroup:
    """Nonsymmorphic "point group" of the diamond and pyrochlore lattices with a
    cubic unit cell of side length 1."""
    return (_inv_group().change_origin([1 / 8, 1 / 8, 1 / 8]) @ Td()).replace(
        unit_cell=np.asarray([[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]])
    )


diamond = pyrochlore = Fd3m



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.hilbert.abstract_hilbert import AbstractHilbert


class ContinuousHilbert(AbstractHilbert):
    """Abstract class for the Hilbert space of particles
    in continuous space.

    This class defines the common interface that
    can be used to interact with particles defined
    in continuous space.
    """

    def __init__(self, domain: tuple[float, ...]):
        """
        Constructs an Hilbert space with continuous degrees of freedom, given specifications of the space.

        This object returns an Hilbert space.

        Args:
            domain: Tuple indicating the maximum of the continuous quantum number(s) in the configurations. Each entry
                in the tuple corresponds to a different physical dimension.
                A particle in a 3D box of size L would take `(L,L,L)`. A rotor model would take e.g. `(2pi,)`.
        """
        self._extent = tuple(domain)

        super().__init__()

    @property
    def domain(self) -> tuple[float, ...]:
        r"""Domain of the continuous variable, specified for each dimension"""
        return self._extent



from netket.sampler.rules import (
    FermionHopRule as _deprecated_FermionHopRule,
)

_deprecations = {
    # June 2024, NetKet 3.13
    "ParticleExchangeRule": (
        "netket.experimental.sampler.rules.ParticleExchangeRule is deprecated: use "
        "netket.sampler.rules.FermionHopRule (netket >= 3.13)",
        _deprecated_FermionHopRule,
    ),
}

from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

__getattr__ = _deprecation_getattr(__name__, _deprecations)
del _deprecation_getattr



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence

import numpy as np

from netket.graph import AbstractGraph, Graph
from netket.hilbert import AbstractHilbert
from netket.utils.types import DType

from ._graph_operator import GraphOperator


class Heisenberg(GraphOperator):
    r"""
    The Heisenberg hamiltonian on a lattice.
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        graph: AbstractGraph,
        J: float | Sequence[float] = 1.0,
        sign_rule: None | bool | Sequence[bool] = None,
        dtype: DType | None = None,
        *,
        acting_on_subspace: None | list[int] | int = None,
    ):
        """
        Constructs an Heisenberg operator given a hilbert space and a graph providing the
        connectivity of the lattice.

        Args:
            hilbert: Hilbert space the operator acts on.
            graph: The graph upon which this hamiltonian is defined.
            J: The strength of the coupling. Default is 1.
               Can pass a sequence of coupling strengths with coloured graphs:
               edges of colour n will have coupling strength J[n]
            sign_rule: If True, Marshal's sign rule will be used. On a bipartite
                lattice, this corresponds to a basis change flipping the Sz direction
                at every odd site of the lattice. For non-bipartite lattices, the
                sign rule cannot be applied. Defaults to True if the lattice is
                bipartite, False otherwise.
                If a sequence of coupling strengths is passed, defaults to False
                and a matching sequence of sign_rule must be specified to override it
            dtype: Data type of the matrix elements.
            acting_on_subspace: Specifies the mapping between nodes of the graph and
                Hilbert space sites, so that graph node :code:`i ∈ [0, ..., graph.n_nodes - 1]`,
                corresponds to :code:`acting_on_subspace[i] ∈ [0, ..., hilbert.n_sites]`.
                Must be a list of length `graph.n_nodes`. Passing a single integer :code:`start`
                is equivalent to :code:`[start, ..., start + graph.n_nodes - 1]`.

        Examples:
         Constructs a ``Heisenberg`` operator for a 1D system.

            >>> import netket as nk
            >>> g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
            >>> hi = nk.hilbert.Spin(s=0.5, total_sz=0, N=g.n_nodes)
            >>> op = nk.operator.Heisenberg(hilbert=hi, graph=g)
            >>> print(op)
            Heisenberg(J=1.0, sign_rule=True; dim=20)
        """
        if isinstance(J, Sequence):
            # check that the number of Js matches the number of colours
            assert len(J) == max(graph.edge_colors) + 1

            if sign_rule is None:
                sign_rule = [False] * len(J)
            else:
                assert len(sign_rule) == len(J)
                for i in range(len(J)):
                    subgraph = Graph(edges=graph.edges(filter_color=i))
                    if sign_rule[i] and not subgraph.is_bipartite():
                        raise ValueError(
                            "sign_rule=True specified for a non-bipartite lattice"
                        )
        else:
            if sign_rule is None:
                sign_rule = graph.is_bipartite()
            elif sign_rule and not graph.is_bipartite():
                raise ValueError("sign_rule=True specified for a non-bipartite lattice")

        self._J = J
        self._sign_rule = sign_rule

        sz_sz = np.array(
            [
                [1, 0, 0, 0],
                [0, -1, 0, 0],
                [0, 0, -1, 0],
                [0, 0, 0, 1],
            ]
        )
        exchange = np.array(
            [
                [0, 0, 0, 0],
                [0, 0, 2, 0],
                [0, 2, 0, 0],
                [0, 0, 0, 0],
            ]
        )

        if isinstance(J, Sequence):
            bond_ops = [
                J[i] * (sz_sz - exchange if sign_rule[i] else sz_sz + exchange)
                for i in range(len(J))
            ]
            bond_ops_colors = list(range(len(J)))
        else:
            bond_ops = [J * (sz_sz - exchange if sign_rule else sz_sz + exchange)]
            bond_ops_colors = []

        super().__init__(
            hilbert,
            graph,
            bond_ops=bond_ops,
            bond_ops_colors=bond_ops_colors,
            dtype=dtype,
            acting_on_subspace=acting_on_subspace,
        )

    @property
    def J(self) -> float:
        """The coupling strength."""
        return self._J

    @property
    def uses_sign_rule(self):
        return self._sign_rule

    def __repr__(self):
        return f"Heisenberg(J={self._J}, sign_rule={self._sign_rule}; dim={self.hilbert.size})"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
from jax import numpy as jnp

from flax import serialization

import netket
from netket import jax as nkjax
from netket.hilbert import DiscreteHilbert
from netket.sampler import Sampler
from netket.stats import Stats
from netket.utils.types import PyTree
from netket.operator import AbstractOperator

from netket.vqs import VariationalMixedState

from netket.vqs.mc import MCState


def apply_diagonal(bare_afun, w, x, *args, **kwargs):
    x = jnp.hstack((x, x))
    return bare_afun(w, x, *args, **kwargs)


class MCMixedState(VariationalMixedState, MCState):
    """Variational State for a Mixed Variational Neural Quantum State.

    The state is sampled according to the provided sampler, and it's diagonal is sampled
    according to another sampler.
    """

    def __init__(
        self,
        sampler,
        model=None,
        *,
        sampler_diag: Sampler | None = None,
        n_samples_diag: int | None = None,
        n_samples_per_rank_diag: int | None = None,
        n_discard_per_chain_diag: int | None = None,
        seed=None,
        sampler_seed: int | None = None,
        variables=None,
        **kwargs,
    ):
        """
        Constructs the MCMixedState.
        Arguments are the same as :class:`MCState`.

        Arguments:
            sampler: The sampler
            model: (Optional) The model. If not provided, you must provide init_fun and apply_fun.
            n_samples: the total number of samples across chains and processes when sampling (default=1000).
            n_samples_per_rank: the total number of samples across chains on one process when sampling. Cannot be
                specified together with n_samples (default=None).
            n_discard_per_chain: number of discarded samples at the beginning of each monte-carlo chain (default=n_samples/10).
            n_samples_diag: the total number of samples across chains and processes when sampling the diagonal
                of the density matrix (default=1000).
            n_samples_per_rank_diag: the total number of samples across chains on one process when sampling the diagonal.
                Cannot be specified together with `n_samples_diag` (default=None).
            n_discard_per_chain_diag: number of discarded samples at the beginning of each monte-carlo chain used when sampling
                the diagonal of the density matrix for observables (default=n_samples_diag/10).
            parameters: Optional PyTree of weights from which to start.
            seed: rng seed used to generate a set of parameters (only if parameters is not passed). Defaults to a random one.
            sampler_seed: rng seed used to initialise the sampler. Defaults to a random one.
            mutable: Name or list of names of mutable arguments. Use it to specify if the model has a state that can change
                during evaluation, but that should not be optimised. See also flax.linen.module.apply documentation
                (default=False)
            init_fun: Function of the signature f(model, shape, rng_key, dtype) -> Optional_state, parameters used to
                initialise the parameters. Defaults to the standard flax initialiser. Only specify if your network has
                a non-standard init method.
            apply_fun: Function of the signature f(model, variables, σ) that should evaluate the model. Defaults to
                `model.apply(variables, σ)`. specify only if your network has a non-standard apply method.
            training_kwargs: a dict containing the optional keyword arguments to be passed to the apply_fun during training.
                Useful for example when you have a batchnorm layer that constructs the average/mean only during training.

        """

        seed, seed_diag = jax.random.split(nkjax.PRNGKey(seed))
        if sampler_seed is None:
            sampler_seed_diag = None
        else:
            sampler_seed, sampler_seed_diag = jax.random.split(
                nkjax.PRNGKey(sampler_seed)
            )

        self._diagonal = None

        hilbert_physical = sampler.hilbert.physical

        super().__init__(
            sampler.hilbert.physical,
            sampler,
            model,
            **kwargs,
            seed=seed,
            sampler_seed=sampler_seed,
            variables=variables,
        )

        if sampler_diag is None:
            sampler_diag = sampler.replace(hilbert=hilbert_physical)

        sampler_diag = sampler_diag.replace(machine_pow=1)

        diagonal_apply_fun = nkjax.HashablePartial(apply_diagonal, self._apply_fun)

        for kw in [
            "n_samples",
            "n_discard_per_chain",
        ]:
            if kw in kwargs:
                kwargs.pop(kw)

        self._diagonal = MCState(
            sampler_diag,
            apply_fun=diagonal_apply_fun,
            n_samples=n_samples_diag,
            n_samples_per_rank=n_samples_per_rank_diag,
            n_discard_per_chain=n_discard_per_chain_diag,
            variables=self.variables,
            seed=seed_diag,
            sampler_seed=sampler_seed_diag,
            **kwargs,
        )

    @property
    def diagonal(self):
        return self._diagonal

    @property
    def sampler_diag(self) -> Sampler:
        """The Monte Carlo sampler used by this Monte Carlo variational state to
        sample the diagonal."""
        return self.diagonal.sampler

    @sampler_diag.setter
    def sampler_diag(self, sampler):
        self.diagonal.sampler = sampler

    @property
    def n_samples_diag(self) -> int:
        """The total number of samples generated at every sampling step
        when sampling the diagonal of this mixed state.
        """
        return self.diagonal.n_samples

    @n_samples_diag.setter
    def n_samples_diag(self, n_samples):
        self.diagonal.n_samples = n_samples

    @property
    def chain_length_diag(self) -> int:
        """
        Length of the markov chain used for sampling the diagonal configurations.

        If running under MPI, the total samples will be n_nodes * chain_length * n_batches.
        """
        return self.diagonal.chain_length

    @chain_length_diag.setter
    def chain_length_diag(self, length: int):
        self.diagonal.chain_length = length

    @property
    def n_discard_per_chain_diag(self) -> int:
        """Number of discarded samples at the beginning of the markov chain used to
        sample the diagonal of this mixed state.
        """
        return self.diagonal.n_discard_per_chain

    @n_discard_per_chain_diag.setter
    def n_discard_per_chain_diag(self, n_discard_per_chain: int | None):
        self.diagonal.n_discard_per_chain = n_discard_per_chain

    @MCState.parameters.setter
    def parameters(self, pars: PyTree):
        MCState.parameters.fset(self, pars)
        if self.diagonal is not None:
            self.diagonal.parameters = pars

    @MCState.model_state.setter
    def model_state(self, state: PyTree):
        MCState.model_state.fset(self, state)
        if self.diagonal is not None:
            self.diagonal.model_state = state

    def reset(self):
        super().reset()
        if self.diagonal is not None:
            self.diagonal.reset()

    def expect_and_grad_operator(
        self, Ô: AbstractOperator, is_hermitian=None
    ) -> tuple[Stats, PyTree]:
        raise NotImplementedError

    def to_matrix(self, normalize: bool = True) -> jax.Array:
        if not isinstance(self.hilbert, DiscreteHilbert):
            raise TypeError("Cannot convert to array a non-discrete Hilbert space.")

        return netket.nn.to_matrix(  # type: ignore[return-value]
            self.hilbert,
            self._apply_fun,
            self.variables,
            normalize=normalize,
            chunk_size=self.chunk_size,
        )

    def __repr__(self):
        return (
            "MCMixedState("
            + f"\n  hilbert = {self.hilbert},"
            + f"\n  sampler = {self.sampler},"
            + f"\n  n_samples = {self.n_samples},"
            + f"\n  n_discard_per_chain = {self.n_discard_per_chain},"
            + f"\n  sampler_state = {self.sampler_state},"
            + f"\n  sampler_diag = {self.sampler_diag},"
            + f"\n  n_samples_diag = {self.n_samples_diag},"
            + f"\n  n_discard_per_chain_diag = {self.n_discard_per_chain_diag},"
            + f"\n  sampler_state_diag = {self.diagonal.sampler_state},"
            + f"\n  n_parameters = {self.n_parameters})"
        )

    def __str__(self):
        return (
            "MCMixedState("
            + f"hilbert = {self.hilbert}, "
            + f"sampler = {self.sampler}, "
            + f"n_samples = {self.n_samples})"
        )


# serialization


def serialize_MCMixedState(vstate):
    state_dict = {
        "variables": serialization.to_state_dict(vstate.variables),
        "sampler_state": serialization.to_state_dict(vstate._sampler_state_previous),
        "diagonal": serialization.to_state_dict(vstate.diagonal),
        "n_samples": vstate.n_samples,
        "n_discard_per_chain": vstate.n_discard_per_chain,
        "chunk_size": vstate.chunk_size,
    }
    return state_dict


def deserialize_MCMixedState(vstate, state_dict):
    import copy

    new_vstate = copy.copy(vstate)
    new_vstate.reset()

    # restore the diagonal first so we can relink the samples
    new_vstate._diagonal = serialization.from_state_dict(
        vstate._diagonal, state_dict["diagonal"]
    )

    new_vstate.variables = jax.tree_util.tree_map(
        jnp.asarray,
        serialization.from_state_dict(vstate.variables, state_dict["variables"]),
    )
    new_vstate.sampler_state = serialization.from_state_dict(
        vstate.sampler_state, state_dict["sampler_state"]
    )

    new_vstate.n_samples = state_dict["n_samples"]
    new_vstate.n_discard_per_chain = state_dict["n_discard_per_chain"]
    new_vstate.chunk_size = state_dict["chunk_size"]

    return new_vstate


serialization.register_serialization_state(
    MCMixedState,
    serialize_MCMixedState,
    deserialize_MCMixedState,
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from collections.abc import Callable

import abc

from netket.utils.types import PyTree, Scalar
from netket.utils import timing, struct
from netket.vqs import VariationalState

from .linear_operator import LinearOperator, SolverT

# Generic signature of a preconditioner function/object

PreconditionerT = Callable[[VariationalState, PyTree, Scalar | None], PyTree]
"""Signature for Gradient preconditioners supported by NetKet drivers."""

LHSConstructorT = Callable[[VariationalState, Scalar | None], LinearOperator]
"""Signature for the constructor of a LinerOperator"""


class IdentityPreconditioner(struct.Pytree):
    """
    A preconditioner that does not transform the gradient.
    """

    def __call__(
        self,
        vstate: VariationalState,
        gradient: PyTree,
        step: Scalar | None = 0,
        *args,
        **kwargs,
    ) -> PyTree:
        return gradient


# For backward compatibility reasons
identity_preconditioner = IdentityPreconditioner()


class AbstractLinearPreconditioner(struct.Pytree, mutable=True):
    """Base class for a Linear Preconditioner solving a system :math:`Sx = F`.

    A LinearPreconditioner modifies the gradient :math:`F` in such a way that the new
    gradient :math:`x` solves the linear system `:math:`Sx=F`. The linear operator
    :math:`S` is constructed from the variational state.

    To subtype this class and provide a concrete implementation, one needs to define
    at least the function

    .. code::

        @dataclass
        class MyLinearPreconditioner(AbstractLinearPreconditioner):

            def lhs_constructor(self, vstate: VariationalState, step: Optional[Scalar] = 0):
                # here the lhs of the system should be constructed, for example by
                # returning the geometric tensor or any other object
                # return vstate.quantum_geometric_tensor()

    """

    solver: SolverT = struct.field(serialize=False)
    """Function used to solve the linear system."""

    solver_restart: bool = False
    """If False uses the last solution of the linear system as a starting point for the solution
    of the next."""

    x0: PyTree | None = None
    """Solution of the last linear system solved."""

    info: Any = struct.field(serialize=False, default=None)
    """Additional information returned by the solver when solving the last linear system."""

    _lhs: LinearOperator = struct.field(serialize=False, default=None)
    """LHS of the last linear system solved."""

    def __init__(self, solver, *, solver_restart=False):
        """
        Constructs the structure holding the parameters for using the
        linear preconditioner.

        Args:
            solver: A callable that solves a linear system of equations.
            solver_restart: If False uses the last solution of the linear
                system as a starting point for the solution of the next
                (default=False).
        """
        self.solver = solver
        self.solver_restart = solver_restart

    @timing.timed
    def __call__(
        self,
        vstate: VariationalState,
        gradient: PyTree,
        step: Scalar | None = None,
        *args,
        **kwargs,
    ) -> PyTree:
        self._lhs = self.lhs_constructor(vstate, step)

        x0 = self.x0 if self.solver_restart else None
        self.x0, self.info = self._lhs.solve(self.solver, gradient, x0=x0)

        return self.x0

    @abc.abstractmethod
    def lhs_constructor(self, vstate: VariationalState, step: Scalar | None = None):
        """
        This method should construct the left hand side of the linear system,
        which should be a linear operator.
        """

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            + f"\n\tsolver          = {self.solver}, "
            + f"\n\tsolver_restart  = {self.solver_restart},"
            + ")"
        )


# This exists for backward compatibility
class LinearPreconditioner(AbstractLinearPreconditioner, mutable=True):
    lhs_constructor: LHSConstructorT
    """Constructor of the LHS of the linear system starting from the variational state."""

    def __init__(
        self,
        lhs_constructor: LHSConstructorT,
        solver: SolverT,
        *,
        solver_restart: bool = False,
    ):
        self._lhs_constructor = lhs_constructor
        self.solver = solver
        self.solver_restart = solver_restart

    def lhs_constructor(self, vstate: VariationalState, step: Scalar | None = None):
        """
        This method does things
        """
        return self._lhs_constructor(vstate, step)

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            + f"\n\tlhs_constructor = {self._lhs_constructor}, "
            + f"\n\tsolver          = {self.solver}, "
            + f"\n\tsolver_restart  = {self.solver_restart},"
            + ")"
        )



from ._common import exec_in_terminal, get_executable_path


def get_global_mpi_info():
    info = {}
    info["mpicc"] = get_executable_path("mpicc")
    info["mpirun"] = get_executable_path("mpirun")
    info["mpiexec"] = get_executable_path("mpiexec")

    if info["mpicc"]:
        info["link_flags"] = get_link_flags(exec_in_terminal([info["mpicc"], "-show"]))
    else:
        info["link_flags"] = ""
    return info


def get_link_flags(flags):
    flags = flags.replace(",", " ").split()
    res = []
    for flag in flags:
        if flag.startswith("-L"):
            res.append(flag)

    return res



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Union
from collections.abc import Iterable
from netket.utils.types import DType, Array

import numpy as np
import jax.numpy as jnp
from numba import jit
from itertools import product
from numbers import Number

from netket import jax as nkjax
from netket.jax import canonicalize_dtypes
from netket.hilbert import Qubit, AbstractHilbert
from netket.utils.numbers import dtype as _dtype, is_scalar

from .._abstract_operator import AbstractOperator
from .._discrete_operator import DiscreteOperator

valid_pauli_regex = re.compile(r"^[XYZI]+$")


def _standardize_matrix_input_type(op):
    """
    Standardize the structural type of operators stored in LocalOperator.

    Eventually, we could also support spmatrices (but some work will be needed.)
    """
    if isinstance(op, list):
        return np.asarray(op)
    else:
        return op


def cast_operator_matrix_dtype(matrix: Array, dtype: DType):
    """
    Changes the dtype of a matrix, without changing the structural type of the object.

    This makes sure that if you pass sparse arrays to a LocalOperator, they remain
    sparse even if you change the dtype
    """
    # must copy
    # return np.asarray(matrix, dtype=dtype)
    return matrix.astype(dtype)


def canonicalize_input(hilbert: AbstractHilbert, operators, weights, *, dtype=None):
    if operators is None:
        operators = []

    # Support single-operator
    if isinstance(operators, str):
        operators = [operators]

    # default weight is 1
    if weights is None:
        weights = True

    if is_scalar(weights):
        weights = [weights for _ in operators]

    if len(weights) != len(operators):
        raise ValueError("weights should have the same length as operators.")

    if hilbert is None:
        if len(operators) > 0:
            hilbert = Qubit(len(operators[0]))
        else:
            raise ValueError(
                "To construct an empty PauliString the hilbert space "
                "must be specified."
            )

    if not np.allclose(hilbert.shape, 2):
        raise ValueError(
            "PauliStrings only work for local hilbert size 2 where PauliMatrices are defined"
        )

    if any(len(op) != hilbert.size for op in operators):
        raise ValueError("Pauli strings have inhomogeneous lengths.")

    consistent = all(bool(valid_pauli_regex.search(op)) for op in operators)
    if not consistent:
        raise ValueError(
            """Operators in string must be one of
            the Pauli operators X,Y,Z, or the identity I"""
        )

    weights = _standardize_matrix_input_type(weights)
    operators = np.asarray(operators, dtype=str)

    operators, weights = _reduce_pauli_string(operators, weights)

    # When there is an odd number of 'Y' in any string, the whole operator must be complex
    op_is_complex = any(s.count("Y") % 2 == 1 for s in operators)

    dtype = canonicalize_dtypes(
        complex if op_is_complex else float, weights, dtype=dtype
    )

    if not nkjax.is_complex_dtype(dtype):
        if op_is_complex:
            raise TypeError("Cannot specify real dtype with an odd number of Y")

        if nkjax.is_complex_dtype(weights.dtype):
            raise TypeError("Cannot specify real dtype with complex weights")

    weights = cast_operator_matrix_dtype(weights, dtype=dtype)

    return hilbert, operators, weights, weights.dtype


class PauliStringsBase(DiscreteOperator):
    """A Hamiltonian consisting of the sum of products of Pauli operators."""

    def __init__(
        self,
        hilbert: AbstractHilbert,
        operators: None | str | list[str] = None,
        weights: None | float | complex | list[float | complex] = None,
        *,
        cutoff: float = 1.0e-10,
        dtype: DType | None = None,
    ):
        """
        Constructs a new ``PauliStrings`` operator given a set of Pauli operators.
        This class has two possible forms for initialization: ``PauliStrings(hilbert, operators, ...)`` or  ``PauliStrings(operators, ...)``.
        When no hilbert argument is passed, the hilbert defaults to Qubit, where the number of qubits is automatically deduced from the operators.

        Args:
           hilbert: A hilbert space, optional (is no ``AbstractHilbert`` is passed, default is Qubit)
           operators (list(string)): A list of Pauli operators in string format, e.g. ['IXX', 'XZI'].
           weights: A list of amplitudes of the corresponding Pauli operator.
           cutoff (float): a cutoff to remove small matrix elements (default = 1e-10)
           dtype: The datatype to use for the matrix elements. Defaults to double precision if
                available.

        Examples:
           Constructs a new ``PauliStrings`` operator X_0*X_1 + 3.*Z_0*Z_1 with both construction schemes.

           >>> import netket as nk
           >>> operators, weights = ['XX','ZZ'], [1,3]
           >>> op = nk.operator.PauliStrings(operators, weights)
           >>> op.hilbert
           Qubit(N=2)
           >>> op.hilbert.size
           2
           >>> hilbert = nk.hilbert.Spin(1/2, 2)
           >>> op = nk.operator.PauliStrings(hilbert, operators, weights)
           >>> op.hilbert
           Spin(s=1/2, N=2, ordering=new)
        """
        if hilbert is None:
            raise ValueError("None-valued hilbert passed.")

        # if first argument is not Hilbert, then shift all arguments by one
        # to support not declaring the Hilbert space
        if not isinstance(hilbert, AbstractHilbert):
            hilbert, operators, weights = None, hilbert, operators

        hilbert, operators, weights, dtype = canonicalize_input(
            hilbert, operators, weights, dtype=dtype
        )

        if not np.isscalar(cutoff) or cutoff < 0:
            raise ValueError("invalid cutoff in PauliStrings.")

        super().__init__(hilbert)

        self._operators = operators
        self._weights = weights
        self._dtype = dtype

        self._cutoff = cutoff

        self._is_hermitian = None

    @property
    def operators(self) -> Iterable[str]:
        return self._operators

    @property
    def weights(self) -> Array:
        return self._weights

    @classmethod
    def identity(cls, hilbert: AbstractHilbert, **kwargs):
        return cls(hilbert, "I" * hilbert.size, **kwargs)

    @classmethod
    def from_openfermion(
        cls,
        hilbert: AbstractHilbert,
        of_qubit_operator=None,  # : "openfermion.ops.QubitOperator" type
        *,
        n_qubits: int | None = None,
    ) -> "PauliStringsBase":
        r"""
        Converts an openfermion QubitOperator into a netket PauliStrings.

        The hilbert first argument can be dropped, see :code:`__init__` for
        details and default value

        Args:
            hilbert: hilbert of the resulting PauliStrings object
            of_qubit_operator: this must be a
                `QubitOperator object <https://quantumai.google/reference/python/openfermion/ops/QubitOperator>`_ .
                More information about those objects can be found in
                `OpenFermion's documentation <https://quantumai.google/reference/python/openfermion>`_
            n_qubits: (optional) total number of qubits in the system, default None means inferring
                it from the QubitOperator. Argument is ignored when hilbert is given.

        """
        from openfermion.ops import QubitOperator

        if hilbert is None:
            raise ValueError("None-valued hilbert passed.")

        if not isinstance(hilbert, AbstractHilbert):
            # if first argument is not Hilbert, then shift all arguments by one
            hilbert, of_qubit_operator = None, hilbert

        if not isinstance(of_qubit_operator, QubitOperator):
            raise NotImplementedError()
        operators = []
        weights = []
        if hilbert is not None:
            # no warning, just overwrite
            n_qubits = hilbert.size
        if n_qubits is None:
            n_qubits = _count_of_locations(of_qubit_operator)
        for operator, weight in of_qubit_operator.terms.items():  # gives dict
            s = ["I"] * n_qubits
            for loc, op in operator:
                assert (
                    loc < n_qubits
                ), f"operator index {loc} is longer than n_qubits={n_qubits}"
                s[loc] = op
            operators.append("".join(s))
            weights.append(weight)

        ps_args = (operators, weights)
        if hilbert is not None:
            ps_args = (hilbert, *ps_args)
        return cls(*ps_args)

    @property
    def dtype(self) -> DType:
        """The dtype of the operator's matrix elements ⟨σ|Ô|σ'⟩."""
        return self._dtype

    @property
    def is_hermitian(self) -> bool:
        """Returns true if this operator is hermitian."""
        if self._is_hermitian is None:
            self._is_hermitian = np.allclose(self._weights.imag, 0.0)
        return self._is_hermitian

    def __repr__(self):
        print_list = []
        for op, w in zip(self.operators, self.weights):
            print_list.append(f"    {op} : {str(w)}")
        s = "{}(hilbert={}, n_strings={}, dtype={}, dict(operators:weights)=\n{}\n)".format(
            type(self).__name__,
            self.hilbert,
            len(self.operators),
            self.dtype,
            ",\n".join(print_list),
        )
        return s

    def copy(self, *, dtype: DType | None = None, cutoff=None):
        """Returns a copy of the operator, while optionally changing the dtype
        of the operator.

        Args:
            dtype: optional dtype
            cutoff: optional override for the cutoff value
        """

        if dtype is None:
            dtype = self.dtype
        if cutoff is None:
            cutoff = self._cutoff

        if not np.can_cast(self.dtype, dtype, casting="same_kind"):
            raise ValueError(f"Cannot cast {self.dtype} to {dtype}")

        new = type(self)(self.hilbert, dtype=dtype)
        new._cutoff = cutoff
        new._operators = self.operators

        if dtype == self.dtype:
            new._weights = self._weights.copy()
        else:
            new._weights = cast_operator_matrix_dtype(self._weights, dtype)

        return new

    def _reset_caches(self):
        self._is_hermitian = None

    def _op__matmul__(self, other):
        if not isinstance(other, PauliStringsBase):
            return NotImplemented
        op = self.copy(
            dtype=jnp.promote_types(self.dtype, _dtype(other)),
            cutoff=min(self._cutoff, other._cutoff),
        )
        return op._op_imatmul_(other)

    def _op_imatmul_(self, other: "PauliStringsBase") -> "PauliStringsBase":
        if not isinstance(other, PauliStringsBase):
            return NotImplemented
        if not self.hilbert == other.hilbert:
            raise ValueError(
                f"Can only multiply identical hilbert spaces (got A@B, A={self.hilbert}, B={other.hilbert})"
            )
        operators, weights = _matmul(
            self.operators,
            self.weights,
            other.operators,
            other.weights,
            dtype=self.dtype,
        )

        self._operators = operators
        self._weights = weights
        self._reset_caches()
        return self

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):
        if isinstance(other, AbstractOperator):
            raise TypeError(
                "To multiply operators use the matrix`@` "
                "multiplication operator `@` instead of the element-wise "
                "multiplication operator `*`.\n\n"
                "For example:\n\n"
                ">>> nk.operator.PauliStrings('XY')@nk.operator.PauliStrings('ZY')"
                "\n\n"
            )
        elif is_scalar(other):
            op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
            return op.__imul__(other)

        return NotImplemented

    def __imul__(self, other):
        if isinstance(other, AbstractOperator):
            raise TypeError(
                "To multiply operators use the matrix`@` "
                "multiplication operator `@` instead of the element-wise "
                "multiplication operator `*`.\n\n"
                "For example:\n\n"
                ">>> nk.operator.PauliStrings('XY')@nk.operator.PauliStrings('ZY')"
                "\n\n"
            )
        elif is_scalar(other):
            if not np.can_cast(_dtype(other), self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot multiply inplace operator of type {type(self)} and "
                    f"dtype {self.dtype} to scalar with dtype {_dtype(other)}"
                )
            other = np.asarray(
                other, dtype=jnp.promote_types(self.dtype, _dtype(other))
            )

            self._weights = self.weights * other
            self._reset_caches()
            return self

        return NotImplemented

    def __sub__(self, other):
        return self.__add__(-other)

    def __rsub__(self, other):
        return other + (-self)

    def __neg__(self):
        return -1 * self

    def __radd__(self, other):
        return self.__add__(other)

    def __isub__(self, other):
        return self.__iadd__(-other)

    def __add__(self, other: Union["PauliStringsBase", Number]):
        op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
        op = op.__iadd__(other)
        return op

    def __iadd__(self, other):
        if isinstance(other, PauliStringsBase):
            if not self.hilbert == other.hilbert:
                raise ValueError(
                    f"Can only add identical hilbert spaces (got A+B, A={self.hilbert}, B={other.hilbert})"
                )

            operators = np.concatenate((self.operators, other.operators))
            weights = np.concatenate((self.weights, other.weights), dtype=self.dtype)
            operators, weights = _reduce_pauli_string(operators, weights)

            self._operators = operators
            self._weights = weights
            self._cutoff = min(self._cutoff, other._cutoff)
            self._reset_caches()
            return self

        elif is_scalar(other):
            if not np.can_cast(_dtype(other), self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator with dtype {_dtype(other)} "
                    f"to operator with dtype {self.dtype}"
                )

            if other != 0.0:
                # adding a constant = adding IIII...III with weight being the constant
                return self.__iadd__(other * self.identity(self.hilbert))
            return self

        raise NotImplementedError


def _count_of_locations(of_qubit_operator):
    """Obtain the number of qubits in the openfermion QubitOperator. Openfermion builds operators from terms that store operators locations.

    Args:
        of_qubit_operator (openfermion.QubitOperator, openfermion.FermionOperator)

    Returns:
        n_qubits (int): number of qubits in the operator, which we can use to create a suitable hilbert space
    """

    # we always start counting from 0, so we only determine the maximum location
    def max_or_default(x):
        x = list(x)
        return max(x) if len(x) > 0 else -1  # -1 is default

    n_qubits = 1 + max_or_default(
        max_or_default(term[0] for term in op) for op in of_qubit_operator.terms.keys()
    )
    return n_qubits


@jit(nopython=True)
def _num_to_pauli(k):
    return ("I", "X", "Y", "Z")[k]


@jit(nopython=True)
def _pauli_to_num(p):
    if p == "X":
        return 1
    elif p == "Y":
        return 2
    elif p == "Z":
        return 3
    elif p == "I":
        return 0
    else:
        raise ValueError("p should be in 'XYZ'")


@jit(nopython=True)
def _levi_term(i, j):
    k = int(6 - i - j)  # i, j, k are permutations of (1,2,3), ijk=0 is already handled
    term = (i - j) * (j - k) * (k - i) / 2
    return _num_to_pauli(k), 1j * term


@jit(nopython=True)
def _apply_pauli_op_reduction(op1, op2):
    if op1 == op2:
        return "I", 1
    elif op1 == "I":
        return op2, 1
    elif op2 == "I":
        return op1, 1
    else:
        n1 = _pauli_to_num(op1)
        n2 = _pauli_to_num(op2)
        pauli, levi_factor = _levi_term(n1, n2)
        return pauli, levi_factor


@jit(nopython=True)
def _split_string(s):
    return [x for x in str(s)]


@jit(nopython=True)
def _make_new_pauli_string(op1, w1, op2, w2):
    """Compute the (symbolic) tensor product of two pauli strings with weights
    Args:
        op1, op2 (str): Pauli strings (e.g. IIXIIXZ).
        w1, w2 (complex): The corresponding weights

    Returns:
        new_op (str): the new pauli string (result of the tensor product)
        new_weight (complex): the weight of the pauli string

    """
    assert len(op1) == len(op2)
    op1 = _split_string(op1)
    op2 = _split_string(op2)
    o_w = [_apply_pauli_op_reduction(a, b) for a, b in zip(op1, op2)]
    new_op = [o[0] for o in o_w]
    new_weights = np.array([o[1] for o in o_w])
    new_op = "".join(new_op)
    new_weight = w1 * w2 * np.prod(new_weights)
    return new_op, new_weight


def _remove_zero_weights(op_arr, w_arr):
    if len(op_arr) == 0:
        return op_arr, w_arr
    idx_nz = ~np.isclose(w_arr, 0)
    if np.any(idx_nz):
        operators = op_arr[idx_nz]
        weights = w_arr[idx_nz]
    else:
        # convention
        operators = np.array(["I" * len(op_arr[0])])
        weights = np.array([0], dtype=w_arr.dtype)
    return operators, weights


def _reduce_pauli_string(op_arr, w_arr):
    """From a list of pauli strings, sum the weights of duplicate strings.
    Args:
        op1, op2 (str): Pauli strings (e.g. IIXIIXZ).
        w1, w2 (complex): The corresponding weights

    Returns:
        new_op (str): the new pauli string (result of the tensor product)
        new_weight (complex): the weight of the pauli string

    """
    operators_unique, idx = np.unique(op_arr, return_inverse=True)
    if len(operators_unique) == len(op_arr):
        # still remove zeros
        return _remove_zero_weights(op_arr, w_arr)
    summed_weights = np.array(
        [np.sum(w_arr[idx == i]) for i in range(len(operators_unique))],
        dtype=w_arr.dtype,
    )
    operators, weights = _remove_zero_weights(operators_unique, summed_weights)
    return operators, weights


def _matmul(op_arr1, w_arr1, op_arr2, w_arr2, *, dtype):
    """(Symbolic) Tensor product of two PauliStrings
    Args:
        op_arr1, op_arr2 (np.array): Arrays operators (strings) in a PauliStrings sum
        w_arr1, w_arr2 (np.array): The corresponding weights of the operators in the sums

    Returns:
        operators (np.array): Array of the resulting operator strings
        new_weight (np.array): Array of the corresponding weights
    """

    operators = []
    weights = []
    for (op1, w1), (op2, w2) in product(zip(op_arr1, w_arr1), zip(op_arr2, w_arr2)):
        # warning: numba always returns complex values, even if we are expecting float.
        op, w = _make_new_pauli_string(op1, w1, op2, w2)
        operators.append(op)
        weights.append(w)
    # so here we recast to the desired dtype
    operators, weights = np.array(operators), np.array(weights)
    # explicit real part to avoid warning
    if not nkjax.is_complex_dtype(dtype):
        weights = weights.real
    weights = weights.astype(dtype)

    operators, weights = _reduce_pauli_string(operators, weights)
    return operators, weights



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from netket.utils import struct


class EarlyStopping(struct.Pytree, mutable=True):
    """A simple callback to stop NetKet if there are no more improvements in the training.
    based on `driver._loss_name`.
    """

    min_delta: float
    """Minimum change in the monitored quantity to qualify as an improvement."""
    min_reldelta: float
    """Minimum relative change in the monitored quantity to qualify as an improvement.

    This behaves similarly to `min_delta` but is more useful for intensive quantities that
    converge to 0, where absolute tolerances might not be effective.
    """
    patience: int | float
    """Number of epochs with no improvement after which training will be stopped."""
    baseline: float | None
    """Baseline value for the monitored quantity. Training will stop if the driver is above the baseline."""
    monitor: str
    """Loss statistic to monitor. Should be one of 'mean', 'variance', 'error_of_mean'."""
    start_from_step: int
    """Number of steps to wait before the callback has any effect."""

    # The quantities below are internal and should not be edited directly
    # by the user

    _best_val: float = np.inf
    """Best value of the loss observed up to this iteration. """
    _best_iter: int
    """Iteration at which the `_best_val` was observed."""
    _best_patience_counter: int
    """Stores the iteration at which we've seen the best loss so far"""

    def __init__(
        self,
        min_delta: float = 0.0,
        min_reldelta: float = 0.0,
        patience: int | float = 0,
        baseline: float | None = None,
        start_from_step: int = 0,
        monitor: str = "mean",
    ):
        """
        Construct an early stopping callback.

        Args:
            min_delta: Minimum change in the monitored quantity to
                qualify as an improvement.
            min_reldelta: Minimum relative change in the monitored
                quantity to qualify as an improvement.
            patience: Number of epochs with no improvement after which
                training will be stopped.
            baseline: Baseline value for the monitored quantity. Training
                will stop if the driver does not drop below the baseline.
            monitor: Loss statistic to monitor. Should be one of
                'mean', 'variance', 'error_of_mean'.
            start_from_step: Number of steps to wait before the callback has
                any effect. Defaults to `0`.
        """
        self.min_delta = min_delta
        self.min_reldelta = min_reldelta
        self.patience = patience
        self.baseline = baseline
        self.monitor = monitor
        self.start_from_step = start_from_step

        self._best_val = np.inf
        self._best_iter = 0
        self._best_patience_counter = 0

    def __call__(self, step, log_data, driver):
        """
        A boolean function that determines whether or not to stop training.

        Args:
            step: An integer corresponding to the step (iteration or epoch) in training.
            log_data: A dictionary containing log data for training.
            driver: A NetKet variational driver.

        Returns:
            A boolean. If True, training continues, else, it does not.
        """

        if step < self.start_from_step:
            return True

        loss = np.real(getattr(log_data[driver._loss_name], self.monitor))

        self._best_patience_counter += 1
        if self._is_improvement(loss, self._best_val):
            self._best_val = loss
            self._best_iter = step

            if self.baseline is None:
                self._best_patience_counter = 0
            elif self._is_improvement(loss, self.baseline):
                # If using baseline, update patience only if we are better than baseline
                self._best_patience_counter = 0

        if self._best_patience_counter > self.patience:
            return False

        return True

    def _is_improvement(self, loss, target):
        # minimal value for absolute and relative improvement
        abs_minval = target - self.min_delta
        rel_minval = target * (1 - self.min_reldelta)
        # minimval value that qualify as an improvement
        minval = min(abs_minval, rel_minval)

        return loss < minval



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    "AbstractSolver",
    "AbstractSolverState",
    "IntegratorState",
    "Integrator",
    "Euler",
    "Heun",
    "Midpoint",
    "RK4",
    "RK12",
    "RK23",
    "RK45",
]

from ._solver import AbstractSolver, AbstractSolverState
from ._integrator_state import IntegratorState
from ._integrator import Integrator
from ._rk._solver import Euler, Heun, Midpoint, RK4, RK12, RK23, RK45

from netket.utils import _hide_submodules

_hide_submodules(__name__)



# Copyright 2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp


from netket.utils.types import Array
from netket.utils import HashableArray, struct
from netket.jax import sort, searchsorted

from .base import HilbertIndex, is_indexable


class LookupTableHilbertIndex(HilbertIndex):
    """Index states according to a pre-defined array containing all possible states.

    Does lookup (states_to_numbers) in log time in the number of states,
    and indexing (numbers_to_states) in constant time.

    The pre-defined array is sorted internally in ascending order (lexicographically).
    Lookup of states not in all_states results in undefined behaviour.
    """

    _all_states: HashableArray = struct.field(pytree_node=False)

    def __init__(self, all_states: Array):
        self._all_states = HashableArray(all_states)

    @property
    def n_states(self) -> int:
        return self.all_states().shape[0]

    @jax.jit
    def numbers_to_states(self, numbers: Array) -> Array:
        return self.all_states()[numbers]

    @jax.jit
    def states_to_numbers(self, states: Array) -> Array:
        return searchsorted(self.all_states(), states)

    @jax.jit
    def all_states(self) -> Array:
        with jax.ensure_compile_time_eval():
            # unpack HashableArray,
            # ensure the states are sorted
            res = sort(jnp.asarray(self._all_states))
        return res

    @property
    def is_indexable(self) -> bool:
        return is_indexable(self.n_states)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Literal, Union

from plum import dispatch, overload, parametric, convert  # noqa: F401


# Todo: deprecated in netket 3.10/august 2023 . To eventually remove.
def __getattr__(name):
    if name in ["TrueT", "FalseT", "Bool"]:
        from netket.utils import warn_deprecation as _warn_deprecation

        _warn_deprecation(
            """
            The variables `nk.utils.dispatch.{TrueT|FalseT|Bool}` are deprecated. Their usages
            should instead be replaced by the following objects:

                `TrueT` should be replaced by `typing.Literal[True]`
                `FalseT` should be replaced by `typing.Literal[False]`
                `Bool` should be replaced by `bool`
            """
        )
        # Deprecated signature-types for True and False
        # TrueT = Literal[True]
        # FalseT = Literal[False]
        # Bool = Union[TrueT, FalseT]
        if name == "TrueT":
            return Literal[True]
        elif name == "FalseT":
            return Literal[False]
        elif name == "Bool":
            return Union[Literal[True], Literal[False]]

    raise AttributeError(f"module {__name__} has no attribute {name}")



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The mc module contains everything that is needed to define the variational
MCState and MCMixedState (two particular kind of VariationalState). This
module also contains all the relevant definitions needed to compute expectation
values and gradients using NetKet-default operators.

In `common.py` some common methods are defined. Those are used both by MCState
and MCMixedState to more easily define expect/expect_and_grad methods.

In `kernels.py` the local-expect kernels are defined that are then passed to the
expect/expect_and_grad methods.
"""

from .common import check_hilbert, get_local_kernel_arguments, get_local_kernel

from .mc_state import MCState
from .mc_mixed_state import MCMixedState



# Copyright 2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import sys
import jax

from .base import ModuleFramework, framework

if TYPE_CHECKING:
    import equinox


# expose jax-stax as a flax module
class EquinoxWrapper:
    def __init__(self, static_module, params_treedef):
        self.static_module = static_module
        self.params_treedef = params_treedef

    def init(self, rng, *args, **kwargs):
        raise RuntimeError("not allowed")

    def apply(
        self,
        variables,
        *args,
        rngs=None,
        method=None,  # noqa: W0613
        mutable=False,
        **kwargs,
    ):
        if mutable is not False:
            raise NotImplementedError("Not implemented")
        if rngs is not None:
            raise NotImplementedError()
        if method is None:
            method = "__call__"

        module = self.recompose(variables)
        fun = getattr(module, method)

        return fun(*args, key=rngs, **kwargs)

    def recompose(self, variables):
        import equinox as eqx

        params = list(variables["params"]["list"])

        params_module = jax.tree.unflatten(self.params_treedef, params)
        module = eqx.combine(params_module, self.static_module)

        return module

    def __repr__(self):
        return f"EquinoxWrapper({self.static_module})"


@framework
class EquinoxFramework(ModuleFramework):
    name: str = "Equinox"

    @property
    def model_contains_parameters(self) -> bool:
        """
        Returns True if the model contains the parameters in the model itself, False
        if the parameters are stored separately.
        """
        return True

    @staticmethod
    def is_loaded() -> bool:
        return "equinox" in sys.modules

    @staticmethod
    def is_my_module(module) -> bool:
        # this will only get called if the module is loaded
        import equinox as eqx  # noqa: E0401

        return isinstance(module, eqx.Module)

    @staticmethod
    def wrap(module: "equinox.Module") -> tuple[dict, EquinoxWrapper]:
        import equinox as eqx

        params, static = eqx.partition(module, eqx.is_array)
        params_leaves, params_treedef = jax.tree.flatten(params)
        variables = {"params": {"list": tuple(params_leaves)}}

        return variables, EquinoxWrapper(static, params_treedef)

    @staticmethod
    def unwrap(
        wrapped_module: EquinoxWrapper, maybe_variables: dict
    ) -> "equinox.Module":
        import equinox as eqx

        params_leaves = maybe_variables["params"]["list"]
        params_module = jax.tree.unflatten(wrapped_module.params_treedef, params_leaves)

        eqx_module = eqx.combine(params_module, wrapped_module.static_module)
        return eqx_module



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from functools import partial

import numpy as np

import jax
from jax import numpy as jnp

from netket import config
from netket.utils.types import PyTree, PRNGKeyT, Array
from netket.utils import struct, mpi
from netket.jax import dtype_real
from netket.jax.sharding import shard_along_axis

from netket.sampler import MetropolisSamplerState, MetropolisSampler
from netket.sampler.rules import LocalRule, ExchangeRule, HamiltonianRule

# Original C++ Implementation
# https://github.com/netket/netket/blob/1e187ae2b9d2aa3f2e53b09fe743e50763d04c9a/Sources/Sampler/metropolis_hastings_pt.hpp
# https://github.com/netket/netket/blob/1e187ae2b9d2aa3f2e53b09fe743e50763d04c9a/Sources/Sampler/metropolis_hastings_pt.cc
# Python port
# https://github.com/netket/netket/blob/87d469aa8c23f71c4838cf09d7ed7b87ff2ea01f/netket/legacy/sampler/numpy/metropolis_hastings_pt.py


class ParallelTemperingSamplerState(MetropolisSamplerState):
    """
    State for the Metropolis Parallel Tempering sampler.

    Contains the usual quantities, as well as statistics about the paralel tempering.
    """

    beta: jnp.ndarray = None
    """The inverse temperatures of the different chains."""

    n_accepted_per_beta: jnp.ndarray = None
    """Total number of moves accepted per beta across all processes since the last reset."""
    beta_0_index: jnp.ndarray = None
    r"""Index of the position of the chain with :math:`\\beta=1`."""
    beta_position: jnp.ndarray = None
    r"""Averaged position of :math:`\\beta=1`."""
    beta_diffusion: jnp.ndarray = None
    """Average variance of the position of :math:`\\beta = 1`."""
    exchange_steps: int = 0
    """Number of exchanges between the different temperatures."""

    def __init__(
        self,
        σ: jnp.ndarray,
        rng: jnp.ndarray,
        rule_state: Any | None,
        beta: jnp.ndarray,
        log_prob: jnp.ndarray | None = None,
    ):
        n_chains, n_replicas = beta.shape

        self.beta = beta
        self.n_accepted_per_beta = jnp.zeros((n_chains, n_replicas), dtype=int)
        self.beta_0_index = jnp.zeros((n_chains,), dtype=int)
        self.beta_position = jnp.zeros((n_chains,), dtype=float)
        self.beta_diffusion = jnp.zeros((n_chains,), dtype=float)
        self.exchange_steps = jnp.zeros((), dtype=int)
        super().__init__(σ, rng=rng, rule_state=rule_state, log_prob=log_prob)
        self.n_accepted_proc = jnp.zeros(
            n_chains, dtype=int
        )  # correct shape is (n_chains,) and not (n_batches,)

    def __repr__(self):
        if self.n_steps > 0:
            acc_string = f"# accepted = {self.n_accepted}/{self.n_steps} ({self.acceptance * 100}%), "
        else:
            acc_string = ""

        text = (
            f"ParallelTemperingSamplerState(# replicas = {self.beta.shape[-1]}, "
            + acc_string
            + f"rng state={self.rng}"
        )
        return text

    @property
    def normalized_diffusion(self):
        r"""
        Average variance of the position of :math:`\\beta = 1`.
        In the ideal case, this quantity should be of order ~[0.2, 1.0]
        """
        diffusion = (
            jnp.sqrt(self.beta_diffusion / self.exchange_steps) / self.beta.shape[-1]
        )
        out, _ = mpi.mpi_mean_jax(diffusion.mean())

        return out

    @property
    def normalized_position(self):
        r"""
        Average position of :math:`\\beta = 1`, normalized and centered around 0.
        """
        position = self.beta_position / float(self.beta.shape[-1] - 1) - 0.5
        out, _ = mpi.mpi_mean_jax(position.mean())

        return out


class ParallelTemperingSampler(MetropolisSampler):
    """
    Metropolis-Hastings with Parallel Tempering sampler.

    This sampler samples an Hilbert space, producing samples off a specific dtype.
    The samples are generated according to a transition rule that must be
    specified.

    The Metropolis Hastings acceptance rule is correted with a temperature.
    """

    n_replicas: int = struct.field(pytree_node=False, default=32)
    """
    The number of replicas evolving with different temperatures for every
    _physical_ markov chain.

    The total number of chains evolved is :code:`n_chains * n_replicas`.
    """
    _beta_sorted: jax.Array = None
    """
    An internal cache for the user-specified betas, sorted.
    """
    _beta_distribution: str = struct.field(pytree_node=False, default="linear")
    """
    An internal for the user-specified distribution of betas.
    """

    def __init__(
        self,
        *args,
        n_replicas: int | None = None,
        betas: str | jax.Array | None = "linear",
        **kwargs,
    ):
        r"""
        :class:`~netket.sampler.ParallelTemperingSampler` is a generic Metropolis-Hastings sampler using
        a transition rule to perform moves in the Markov Chain.
        The transition kernel is used to generate
        a proposed state :math:`s^\prime`, starting from the current state :math:`s`.
        The move is accepted with probability

        .. math::
            A(s\rightarrow s^\prime) = \mathrm{min}\left (1,\frac{P(s^\prime)}{P(s)} e^{L(s,s^\prime)} \right),

        where the probability being sampled from is :math:`P(s)=β|M(s)|^p`. Here :math:`M(s)` is a
        user-provided function (the machine), :math:`p` is also user-provided with default value :math:`p=2`,
        :math:`β` is the temperature of the Markov Chain and :math:`L(s,s^\prime)` is a suitable correcting factor
        computed by the transition kernel.


        Args:
            hilbert: The hilbert space to sample
            rule: A `MetropolisRule` to generate random transitions from a given state as
                    well as uniform random states.
            n_replicas: The number of different temperatures β for the sampling, must be even.
                    (default : 32).
            betas: (Optional) Distribution or list of values of the temperatures β.
                    For the distribution, possibility between "linear" for a linear distribution
                    and "log" for a logarithmic one.
                    For the explicit list of values, the length must be even and the value β=1 must
                    obligatory be an element of betas, all other temperatures must be in (0,1].
                    (default : "lin", i.e. linear distribution between (0,1]).
            n_chains: The number of Markov Chain to be run in parallel on a single process.
            sweep_size: The number of exchanges that compose a single sweep.
                    If None, sweep_size is equal to the number of degrees of freedom being sampled
                    (the size of the input vector s to the machine).
            n_chains: The number of batches of the states to sample (default = 8)
            machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
            dtype: The dtype of the states sampled (default = np.float32).
        """
        if isinstance(betas, str):
            betas = betas.lower()
            if betas not in ["linear", "lin", "logarithmic", "log"]:
                raise ValueError(
                    f"""
                    To initialize the temperatures with a string, you must choose between
                    "lin" for a linear distribution and "log" for a logarithmic distribution.
                    Instead got "{betas}".
                    """
                )
            if n_replicas is None:
                n_replicas = 32

            # truncate to last 3
            beta_distribution = betas[:3]
            betas = None
        elif isinstance(betas, Array) or isinstance(betas, list):
            # TODO: this defaults to float32/64 depending on what is enabled.
            # Might need to think about a better default here.
            betas = jnp.array(betas, dtype=float)
            if betas.ndim != 1:
                raise ValueError("betas must have exactly 1 dimension.")
            if n_replicas is not None:
                raise ValueError(
                    """
                    Cannot specify the list of betas and n_replicas at the same time.
                    The number of replicas will be inferred automatically from the length
                    of the vector of inverse temperatures.
                    """
                )
            # we need beta[0] = 1, so we sort and check that the temperatures are valid
            betas = jnp.sort(betas, descending=True)
            if not (jnp.isclose(betas[0], 1) and betas[-1] > 0):
                raise ValueError(
                    rf"""The values for beta should be in (0,1] and obligatory contain beta=1, instead got [{jnp.min(betas):.2f},{jnp.max(betas):.8f}]."""
                )

            beta_distribution = "custom"
            n_replicas = betas.shape[-1]
        else:
            raise TypeError(
                "`betas` must be a string or a vector of inverse temperatures."
            )

        # verify the number of replicas
        if not (
            isinstance(n_replicas, int)
            and n_replicas > 0
            and np.mod(n_replicas, 2) == 0
        ):
            raise ValueError(
                "n_replicas (or the length of `betas`) must be an even integer > 0."
            )

        self.n_replicas = n_replicas
        self._beta_sorted = betas
        self._beta_distribution = beta_distribution

        super().__init__(*args, **kwargs)

    @property
    def sorted_betas(self):
        """
        The sorted values of the temperatures for each _physical_ markov chain.
        The first value is β = 1 and is the _physical_ temperature.
        """

        if self._beta_sorted is not None:
            return self._beta_sorted
        else:
            if self._beta_distribution == "lin":
                return (
                    1.0
                    - jnp.arange(self.n_replicas, dtype=jnp.float64) / self.n_replicas
                )
            elif self._beta_distribution == "log":
                return -jnp.log(
                    jnp.arange(1, self.n_replicas + 1, dtype=jnp.float64)
                    / (self.n_replicas + 1)
                ) / jnp.log(self.n_replicas + 1)
            else:
                raise NotImplementedError(f"distribution: {self._beta_distribution}")

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            + f"\n  hilbert = {self.hilbert},"
            + f"\n  rule = {self.rule},"
            + f"\n  n_chains = {self.n_chains},"
            + f"\n  n_replicas = {self.n_replicas},"
            + f"\n  beta_distribution = {self._beta_distribution},"
            + f"\n  sweep_size = {self.sweep_size},"
            + f"\n  reset_chains = {self.reset_chains},"
            + f"\n  machine_power = {self.machine_pow},"
            + f"\n  dtype = {self.dtype}"
            + ")"
        )

    @property
    def n_batches(self) -> int:
        r"""
        The batch size of the configuration $\sigma$ used by this sampler on this
        jax process.

        If you are not using MPI, this is equal to `n_chains * n_replicas`, but if
        you are using MPI this is equal to `n_chains_per_rank * n_replicas`.
        """
        if config.netket_experimental_sharding:
            n_batches = self.n_chains
        else:
            n_batches, remainder = divmod(self.n_chains, mpi.n_nodes)
            if remainder != 0:
                raise RuntimeError(
                    "The number of chains is not a multiple of the number of mpi ranks"
                )
        return n_batches * self.n_replicas

    @partial(jax.jit, static_argnums=1)
    def _init_state(
        self, machine, parameters: PyTree, key: PRNGKeyT
    ) -> ParallelTemperingSamplerState:
        key_state, key_rule, rng = jax.random.split(key, 3)
        rule_state = self.rule.init_state(self, machine, parameters, key_rule)
        σ = self.rule.random_state(self, machine, parameters, rule_state, rng)
        σ = shard_along_axis(σ, axis=0)

        output_dtype = jax.eval_shape(machine.apply, parameters, σ).dtype
        log_prob = jnp.full((self.n_batches,), -jnp.inf, dtype=dtype_real(output_dtype))
        log_prob = shard_along_axis(log_prob, axis=0)

        beta = jnp.tile(self.sorted_betas, (self.n_batches // self.n_replicas, 1))

        return ParallelTemperingSamplerState(
            σ=σ,
            log_prob=log_prob,
            rng=key_state,
            rule_state=rule_state,
            beta=beta,
        )

    @partial(jax.jit, static_argnums=1)
    def _reset(self, machine, parameters: PyTree, state: ParallelTemperingSamplerState):
        state = super()._reset(machine, parameters, state)
        return state.replace(
            n_accepted_per_beta=jnp.zeros_like(state.n_accepted_per_beta),
            beta_position=jnp.zeros_like(state.beta_position),
            beta_diffusion=jnp.zeros_like(state.beta_diffusion),
            exchange_steps=jnp.zeros_like(state.exchange_steps),
            # beta=beta,
            # beta_0_index=jnp.zeros((self.n_chains,), dtype=jnp.int64),
        )

    def _sample_next(
        self, machine, parameters: PyTree, state: ParallelTemperingSamplerState
    ):
        def loop_body(i, s):
            # 1 to propagate for next iteration, 1 for uniform rng and n_chains for transition kernel
            s["key"], key1, key2, key3, key4 = jax.random.split(s["key"], 5)

            # def cbi(data):
            #    i, beta = data
            #    print("sweep #", i, " for beta=\n", beta)
            #    return beta
            #
            # beta = hcb.call(
            #   cbi,
            #   (i, s["beta"]),
            #   result_shape=jax.ShapeDtypeStruct(s["beta"].shape, s["beta"].dtype),
            # )

            beta = s["beta"]

            ## Usual Metropolis sampling
            σp, log_prob_correction = self.rule.transition(
                self, machine, parameters, state, key1, s["σ"]
            )
            proposal_log_prob = self.machine_pow * machine.apply(parameters, σp).real

            uniform = jax.random.uniform(key2, shape=(self.n_batches,))
            if log_prob_correction is not None:
                do_accept = uniform < jnp.exp(
                    beta.reshape((-1,))
                    * (proposal_log_prob - s["log_prob"] + log_prob_correction)
                )
            else:
                do_accept = uniform < jnp.exp(
                    beta.reshape((-1,)) * (proposal_log_prob - s["log_prob"])
                )

            # do_accept must match ndim of proposal and state (which is 2)
            s["σ"] = jnp.where(do_accept.reshape(-1, 1), σp, s["σ"])
            n_accepted_per_beta = s["n_accepted_per_beta"] + do_accept.reshape(
                (self.n_batches // self.n_replicas, self.n_replicas)
            )

            s["log_prob"] = jax.numpy.where(
                do_accept.reshape(-1), proposal_log_prob, s["log_prob"]
            )

            ## exchange betas

            # randomly decide if every set of replicas should be swapped in even or odd order
            swap_order = jax.random.randint(
                key3,
                minval=0,
                maxval=2,
                shape=(self.n_batches // self.n_replicas,),
            )  # 0 or 1

            # indices of even swapped elements (per-row)
            idxs = jnp.arange(0, self.n_replicas, 2).reshape(
                (1, -1)
            ) + swap_order.reshape((-1, 1))
            # indices off odd swapped elements (per-row)
            inn = (idxs + 1) % self.n_replicas

            # for every rows of the input, swap elements at idxs with elements at inn
            @partial(jax.vmap, in_axes=(0, 0, 0), out_axes=0)
            def swap_rows(beta_row, idxs, inn):
                proposed_beta = beta_row.at[idxs].set(
                    beta_row[inn], unique_indices=True, indices_are_sorted=True
                )
                proposed_beta = proposed_beta.at[inn].set(
                    beta_row[idxs], unique_indices=True, indices_are_sorted=False
                )
                return proposed_beta

            proposed_beta = swap_rows(beta, idxs, inn)

            @partial(jax.vmap, in_axes=(0, 0, 0), out_axes=0)
            def compute_proposed_prob(prob, idxs, inn):
                # prob[idxs] = (beta_i - beta_j) log psi(x_i)
                # prob[inn] = (beta_j - beta_i) log psi(x_j)
                # so we have to add the log probabilities to get the right acceptance
                prob_rescaled = prob[idxs] + prob[inn]
                return prob_rescaled

            # compute the probability of the swaps
            log_prob = (proposed_beta - s["beta"]) * s["log_prob"].reshape(
                (self.n_batches // self.n_replicas, self.n_replicas)
            )

            prob_rescaled = jnp.exp(compute_proposed_prob(log_prob, idxs, inn))

            uniform = jax.random.uniform(
                key4,
                shape=(
                    self.n_batches // self.n_replicas,
                    self.n_replicas // 2,
                ),
            )

            # decide where to swap
            do_swap = uniform < prob_rescaled

            do_swap = jnp.dstack((do_swap, do_swap)).reshape(
                (-1, self.n_replicas)
            )  # concat along last dimension

            # roll if swap_order is odd
            do_swap = jax.vmap(jnp.where, in_axes=(0, 0, 0), out_axes=0)(
                swap_order == 0, do_swap, jnp.roll(do_swap, 1, axis=-1)
            )

            # Do the swap where it has to be done
            new_beta = jax.numpy.where(do_swap, proposed_beta, beta)
            s["beta"] = new_beta

            swap_order = swap_order.reshape(-1)

            beta_0_moved = jax.vmap(jnp.take)(
                do_swap, s["beta_0_index"]
            )  # flag saying if beta_0 should move
            proposed_beta_0_index = jnp.mod(
                s["beta_0_index"]
                + (-2 * jnp.mod(swap_order, 2) + 1)
                * (-2 * jnp.mod(s["beta_0_index"], 2) + 1),
                self.n_replicas,
            )

            s["beta_0_index"] = jnp.where(
                beta_0_moved, proposed_beta_0_index, s["beta_0_index"]
            )

            # swap acceptances
            swapped_n_accepted_per_beta = swap_rows(n_accepted_per_beta, idxs, inn)
            s["n_accepted_per_beta"] = jax.numpy.where(
                do_swap,
                swapped_n_accepted_per_beta,
                n_accepted_per_beta,
            )

            # Update statistics to compute diffusion coefficient of replicas
            # Total exchange steps performed
            s["exchange_steps"] += 1
            delta = s["beta_0_index"] - s["beta_position"]
            s["beta_position"] = s["beta_position"] + delta / s["exchange_steps"]
            delta2 = s["beta_0_index"] - s["beta_position"]
            s["beta_diffusion"] = s["beta_diffusion"] + delta * delta2

            return s

        s = {
            "key": state.rng,
            "σ": state.σ,
            # Log prob is already computed in reset, so don't recompute it.
            # "log_prob": self.machine_pow * apply_machine(parameters, state.σ).real,
            "log_prob": state.log_prob,
            "beta": state.beta,
            # for logging
            "beta_0_index": state.beta_0_index,
            "n_accepted_per_beta": state.n_accepted_per_beta,
            "beta_position": state.beta_position,
            "beta_diffusion": state.beta_diffusion,
            "exchange_steps": state.exchange_steps,
        }
        s = jax.lax.fori_loop(0, self.sweep_size, loop_body, s)

        n_accepted_proc = jax.vmap(jnp.take)(
            s["n_accepted_per_beta"], s["beta_0_index"]
        )

        new_state = state.replace(
            rng=s["key"],
            σ=s["σ"],
            log_prob=s["log_prob"],
            n_steps_proc=state.n_steps_proc
            + self.sweep_size * self.n_batches // self.n_replicas,
            beta=s["beta"],
            beta_0_index=s["beta_0_index"],
            beta_position=s["beta_position"],
            beta_diffusion=s["beta_diffusion"],
            exchange_steps=s["exchange_steps"],
            n_accepted_per_beta=s["n_accepted_per_beta"],
            n_accepted_proc=n_accepted_proc,
        )
        σ_flat = new_state.σ
        σ = σ_flat.reshape((-1, self.n_replicas, σ_flat.shape[-1]))
        σ_new = jnp.take_along_axis(σ, s["beta_0_index"][:, None, None], axis=1)
        σ_new = jax.lax.collapse(σ_new, 0, 2)  # remove dummy replica dim

        log_prob = new_state.log_prob.reshape((-1, self.n_replicas))
        log_prob_new = jnp.take_along_axis(log_prob, s["beta_0_index"][:, None], axis=1)
        log_prob_new = jax.lax.collapse(log_prob_new, 0, 2)  # remove dummy replica dim

        return new_state, (σ_new, log_prob_new)


def ParallelTemperingLocal(hilbert, *args, **kwargs):
    r"""
    Sampler acting on one local degree of freedom.

    This sampler acts locally only on one local degree of freedom :math:`s_i`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s_N`,
    where :math:`s^\prime_i \neq s_i`.

    The transition probability associated to this
    sampler can be decomposed into two steps:

    1. One of the site indices :math:`i = 1\dots N` is chosen
    with uniform probability.

    2. Among all the possible (:math:`m`) values that :math:`s_i` can take,
    one of them is chosen with uniform probability.

    For example, in the case of spin :math:`1/2` particles, :math:`m=2`
    and the possible local values are :math:`s_i = -1,+1`.
    In this case then :class:`MetropolisLocal` is equivalent to flipping a random spin.

    In the case of bosons, with occupation numbers
    :math:`s_i = 0, 1, \dots n_{\mathrm{max}}`, :class:`MetropolisLocal`
    would pick a random local occupation number uniformly between :math:`0`
    and :math:`n_{\mathrm{max}}`.

    Args:
        hilbert: The hilbert space to sample
        n_chains: The number of Markov Chain to be run in parallel on a single process.
        sweep_size: The number of exchanges that compose a single sweep.
                If None, sweep_size is equal to the number of degrees of freedom being sampled
                (the size of the input vector s to the machine).
        n_chains: The number of batches of the states to sample (default = 8)
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the states sampled (default = np.float32).
    """
    return ParallelTemperingSampler(hilbert, LocalRule(), *args, **kwargs)


def ParallelTemperingExchange(
    hilbert, *args, clusters=None, graph=None, d_max=1, **kwargs
):
    r"""
    This sampler acts locally only on two local degree of freedom :math:`s_i` and :math:`s_j`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s^\prime_j \dots s_N`,
    where in general :math:`s^\prime_i \neq s_i` and :math:`s^\prime_j \neq s_j`.
    The sites :math:`i` and :math:`j` are also chosen to be within a maximum graph
    distance of :math:`d_{\mathrm{max}}`.

    The transition probability associated to this sampler can
    be decomposed into two steps:

    1. A pair of indices :math:`i,j = 1\dots N`, and such
       that :math:`\mathrm{dist}(i,j) \leq d_{\mathrm{max}}`,
       is chosen with uniform probability.

    2. The sites are exchanged, i.e. :math:`s^\prime_i = s_j` and :math:`s^\prime_j = s_i`.

    Notice that this sampling method generates random permutations of the quantum
    numbers, thus global quantities such as the sum of the local quantum numbers
    are conserved during the sampling.
    This scheme should be used then only when sampling in a
    region where :math:`\sum_i s_i = \mathrm{constant}` is needed,
    otherwise the sampling would be strongly not ergodic.

    Args:
        hilbert: The hilbert space to sample
        d_max: The maximum graph distance allowed for exchanges.
        n_chains: The number of Markov Chain to be run in parallel on a single process.
        sweep_size: The number of exchanges that compose a single sweep.
                If None, sweep_size is equal to the number of degrees of freedom being sampled
                (the size of the input vector s to the machine).
        n_chains: The number of batches of the states to sample (default = 8)
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the states sampled (default = np.float32).


    Examples:
          Sampling from a RBM machine in a 1D lattice of spin 1/2, using
          nearest-neighbours exchanges.

          >>> import netket as nk
          >>>
          >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
          >>> hi=nk.hilbert.Spin(s=0.5, N=g.n_nodes)
          >>>
          >>> # Construct a MetropolisExchange Sampler
          >>> sa = nk.sampler.ParallelTemperingExchange(hi, graph=g)
          >>> print(sa)
          ParallelTemperingSampler(rule = ExchangeRule(# of clusters: 200), n_chains = 16, sweep_size = 100, reset_chains = False, machine_power = 2, dtype = int8)

    """
    rule = ExchangeRule(clusters=clusters, graph=graph, d_max=d_max)
    return ParallelTemperingSampler(hilbert, rule, *args, **kwargs)


def ParallelTemperingHamiltonian(hilbert, hamiltonian, *args, **kwargs):
    r"""
    Sampling based on the off-diagonal elements of a Hamiltonian (or a generic Operator).
    In this case, the transition matrix is taken to be:

    .. math::
       T( \mathbf{s} \rightarrow \mathbf{s}^\prime) = \frac{1}{\mathcal{N}(\mathbf{s})}\theta(|H_{\mathbf{s},\mathbf{s}^\prime}|),

    where :math:`\theta(x)` is the Heaviside step function, and :math:`\mathcal{N}(\mathbf{s})`
    is a state-dependent normalization.
    The effect of this transition probability is then to connect (with uniform probability)
    a given state :math:`\mathbf{s}` to all those states :math:`\mathbf{s}^\prime` for which the Hamiltonian has
    finite matrix elements.
    Notice that this sampler preserves by construction all the symmetries
    of the Hamiltonian. This is in generally not true for the local samplers instead.

    Args:
       machine: A machine :math:`\Psi(s)` used for the sampling.
                The probability distribution being sampled
                from is :math:`F(\Psi(s))`, where the function
                :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
       hamiltonian: The operator used to perform off-diagonal transition.
       n_chains: The number of Markov Chain to be run in parallel on a single process.
       sweep_size: The number of exchanges that compose a single sweep.
                   If None, sweep_size is equal to the number of degrees of freedom (n_visible).


    Examples:
       Sampling from a RBM machine in a 1D lattice of spin 1/2

       >>> import netket as nk
       >>>
       >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
       >>> hi=nk.hilbert.Spin(s=0.5, N=g.n_nodes)
       >>>
       >>> # Transverse-field Ising Hamiltonian
       >>> ha = nk.operator.Ising(hilbert=hi, h=1.0, graph=g)
       >>>
       >>> # Construct a MetropolisExchange Sampler
       >>> sa = nk.sampler.ParallelTemperingHamiltonian(hi, hamiltonian=ha)
       >>> print(sa)
       ParallelTemperingSampler(rule = HamiltonianRuleNumba(operator=Ising(J=1.0, h=1.0; dim=100)), n_chains = 16, sweep_size = 100, reset_chains = False, machine_power = 2, dtype = int8)

    """
    rule = HamiltonianRule(hamiltonian)
    return ParallelTemperingSampler(hilbert, rule, *args, **kwargs)



# Copyright 2021-2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, SupportsFloat, SupportsComplex, TYPE_CHECKING

import numbers


import numpy as np
import jax.numpy as jnp
from scipy.sparse import issparse

from netket.hilbert import AbstractHilbert
from netket.utils.types import DType, Array
from netket.utils.numbers import dtype as _dtype, is_scalar

from .._discrete_operator import DiscreteOperator
from .._lazy import Transpose

from .helpers import (
    canonicalize_input,
    _multiply_operators,
    cast_operator_matrix_dtype,
)
from .convert import local_operators_to_pauli_strings

if TYPE_CHECKING:
    from .._pauli_strings import PauliStrings


def is_hermitian(a: np.ndarray, rtol=1e-05, atol=1e-08) -> bool:
    if issparse(a):
        return np.allclose(a.todense(), a.T.conj().todense(), rtol=rtol, atol=atol)
    else:
        return np.allclose(a, a.T.conj(), rtol=rtol, atol=atol)


def _is_sorted(a):
    for i in range(len(a) - 1):
        if a[i + 1] < a[i]:
            return False
    return True


class LocalOperatorBase(DiscreteOperator):
    """A custom local operator. This is a sum of an arbitrary number of operators
    acting locally on a limited set of k quantum numbers (i.e. k-local,
    in the quantum information sense).
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        operators: list[Array] | Array = [],
        acting_on: list[int] | list[list[int]] = [],
        constant: SupportsComplex | SupportsFloat = 0,
        dtype: DType | None = None,
        *,
        mel_cutoff: float = 1.0e-10,
    ):
        r"""
        Constructs a new ``LocalOperator`` given a hilbert space and (if
        specified) a constant level shift.

        Args:
           hilbert: Hilbert space the operator acts on.
           operators: A list of operators, in matrix form. Supports numpy dense or scipy
           sparse format
           acting_on: A list of list of sites, which the corresponding operators act on. This
                should be constructed such that :code:`operators[i]` acts on the sites :code:`acting_on[i]`.
                If operators is not a list of operators, acting_on should just be the list of
                corresponding sites.
           constant: Constant diagonal shift of the operator, equivalent to
                :math:`+\text{c}\hat{I}`. Default is 0.0.
           dtype: The datatype to use for the matrix elements. Defaults to double precision if
                available.
           mel_cutoff (float): a cutoff to remove small matrix elements (default = 1e-10)

        Examples:
           Constructs a ``LocalOperator`` without any operators.

           >>> from netket.hilbert import Spin
           >>> from netket.operator import LocalOperator
           >>> hi = Spin(0.5)**20
           >>> empty_hat = LocalOperator(hi)
           >>> print(len(empty_hat.acting_on))
           0
        """
        super().__init__(hilbert)
        self.mel_cutoff = mel_cutoff
        self._initialized = None
        self._is_hermitian = None

        # Canonicalize input. From now on input is guaranteed to be in canonical order
        operators, acting_on, dtype = canonicalize_input(
            self.hilbert, operators, acting_on, constant, dtype=dtype
        )
        self._dtype = dtype
        self._constant = np.array(constant, dtype=dtype)

        self._operators_dict = {}
        for op, aon in zip(operators, acting_on):
            self._add_operator(aon, op)

    def _add_operator(self, acting_on: tuple, operator: Array):
        """
        Adds an operator acting on a subset of sites.

        Does not modify in-place the operators themselves which are treated as
        immutables.
        """
        assert isinstance(acting_on, tuple)
        # acting_on_key = tuple(acting_on)
        if acting_on in self._operators_dict:
            operator = self._operators_dict[acting_on] + operator

        self._operators_dict[acting_on] = operator

    @property
    def operators(self) -> list[np.ndarray]:
        """List of the matrices of the operators encoded in this Local Operator.
        Returns a copy.
        """
        return list(self._operators_dict.values())

    @property
    def _operators(self) -> list[np.ndarray]:
        return self.operators

    @property
    def acting_on(self) -> list[list[int]]:
        """List containing the list of the sites on which every operator acts.

        Every operator `self.operators[i]` acts on the sites `self.acting_on[i]`
        """
        return list(self._operators_dict.keys())

    @property
    def n_operators(self) -> int:
        """The total number of operators that were summed upon to build this operator.

        This excludes an optional identity term that is tracked by
        :attr:`netket.operator.LocalOperator.constant`.
        """
        return len(self._operators_dict)

    @property
    def dtype(self) -> DType:
        """DType of the matrix elements of this operator."""
        return self._dtype

    @property
    def size(self) -> int:
        return self._size

    @property
    def is_hermitian(self) -> bool:
        """Returns true if this operator is hermitian."""
        # TODO: (VolodyaCO) I guess that if we have an operator with diagonal elements equal to 1j*C+Y, some complex constant, and
        # self._constant=-1j*C, then the actual diagonal would be Y. How do we check hermiticity taking into account the diagonal
        # elements as well as the self._constant? For the moment I just check hermiticity of the added constant, which must be real.
        if self._is_hermitian is None:
            self._is_hermitian = all(map(is_hermitian, self.operators)) and np.isreal(
                self._constant
            )

        return self._is_hermitian

    @property
    def mel_cutoff(self) -> float:
        r"""float: The cutoff for matrix elements.
        Only matrix elements such that abs(O(i,i))>mel_cutoff
        are considered"""
        return self._mel_cutoff

    @mel_cutoff.setter
    def mel_cutoff(self, mel_cutoff):
        self._mel_cutoff = mel_cutoff
        assert self.mel_cutoff >= 0

    @property
    def constant(self) -> numbers.Number:
        """
        A constant multiplying the identity added to the diagonal of the operator.
        """
        return self._constant

    def to_pauli_strings(self, **kwargs) -> "PauliStrings":  # noqa: F821
        """Convert to PauliStrings object"""
        return local_operators_to_pauli_strings(
            self.hilbert,
            self.operators,
            self.acting_on,
            self.constant,
            self.dtype,
            **kwargs,
        )

    def copy(self, *, dtype: DType | None = None, _cls=None):
        """Returns a copy of the operator, while optionally changing the dtype
        of the operator.

        Args:
            dtype: optional dtype

        Internal args:
            _cls: used to specify the target class
        """

        if dtype is None:
            dtype = self.dtype

        if not np.can_cast(self.dtype, dtype, casting="same_kind"):
            raise ValueError(f"Cannot cast {self.dtype} to {dtype}")

        if _cls is None:
            _cls = type(self)

        new = _cls(self.hilbert, constant=self.constant, dtype=dtype)
        new.mel_cutoff = self.mel_cutoff

        if dtype == self.dtype:
            new._operators_dict = self._operators_dict.copy()
        else:
            new._operators_dict = {
                aon: cast_operator_matrix_dtype(op, dtype)
                for aon, op in self._operators_dict.items()
            }

        return new

    def transpose(self, *, concrete=False):
        r"""LocalOperator: Returns the transpose of this operator."""
        if concrete:
            new = self.copy()
            for aon in new._operators_dict.keys():
                new._operators_dict[aon] = new._operators_dict[aon].transpose()
            return new
        else:
            return Transpose(self)

    def conjugate(self, *, concrete=False):
        r"""LocalOperator: Returns the complex conjugate of this operator."""
        new = self.copy()
        for aon in new._operators_dict.keys():
            new._operators_dict[aon] = new._operators_dict[aon].copy().conjugate()
        return new

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __isub__(self, other):
        return self.__iadd__(-other)

    def __neg__(self):
        return -1 * self

    def __add__(self, other: Union["LocalOperatorBase", numbers.Number]):
        op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
        op = op.__iadd__(other)
        return op

    def __iadd__(self, other):
        if isinstance(other, LocalOperatorBase):
            if self.hilbert != other.hilbert:
                return NotImplemented

            if not np.can_cast(other.dtype, self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator with dtype {other.dtype} "
                    f"to operator with dtype {self.dtype}"
                )

            self.mel_cutoff = min(other.mel_cutoff, self.mel_cutoff)
            self._constant += other.constant.item()
            for aon, op in other._operators_dict.items():
                self._add_operator(aon, op)

            self._reset_caches()
            return self
        if is_scalar(other):
            if not np.can_cast(type(other), self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator with dtype {type(other)} "
                    f"to operator with dtype {self.dtype}"
                )

            self._reset_caches()
            self._constant += other
            return self
        return NotImplemented

    def __truediv__(self, other):
        if not is_scalar(other):
            raise TypeError("Only division by a scalar number is supported.")

        if other == 0:
            raise ValueError("Dividing by 0")
        return self.__mul__(1.0 / other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):
        if isinstance(other, DiscreteOperator):
            op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
            return op.__imatmul__(other)
        elif is_scalar(other):
            op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
            return op.__imul__(other)
        return NotImplemented

    def __imul__(self, other):
        if isinstance(other, DiscreteOperator):
            return self.__imatmul__(other)
        elif is_scalar(other):
            if not np.can_cast(_dtype(other), self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator of type {type(other)} and "
                    f"dtype {_dtype(other)} to operator with dtype {self.dtype}"
                )
            other = np.asarray(
                other, dtype=jnp.promote_types(self.dtype, _dtype(other))
            )

            self._constant *= other
            if np.abs(other) <= self.mel_cutoff:
                self._operators_dict = {}
            else:
                for key in self._operators_dict:
                    self._operators_dict[key] = other * self._operators_dict[key]

            self._reset_caches()
            return self

        return NotImplemented

    def __imatmul__(self, other):
        if not isinstance(other, LocalOperatorBase):
            return NotImplemented

        if not np.can_cast(other.dtype, self.dtype, casting="same_kind"):
            raise ValueError(
                f"Cannot add inplace operator with dtype {type(other)} to operator with dtype {self.dtype}"
            )

        return self._op_imatmul_(other)

    def _op__matmul__(self, other: "LocalOperatorBase") -> "LocalOperatorBase":
        if not isinstance(other, LocalOperatorBase):
            return NotImplemented
        op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
        return op._op_imatmul_(other)

    def _op_imatmul_(self, other: "LocalOperatorBase") -> "LocalOperatorBase":
        if not isinstance(other, LocalOperatorBase):
            return NotImplemented

        self.mel_cutoff = min(other.mel_cutoff, self.mel_cutoff)

        # (α + ∑ᵢAᵢ)(β + ∑ᵢBᵢ) =
        # = αβ + α ∑ᵢBᵢ + β ∑ᵢAᵢ + ∑ᵢⱼAᵢBⱼ
        # = β(α + ∑ᵢAᵢ) + α ∑ᵢBᵢ + ∑ᵢⱼAᵢBⱼ

        α = self.constant.item()
        β = other.constant.item()
        # copy A dict because it is modified inplace in __imul__(β) and add_operators
        A_op_dict = self._operators_dict.copy()
        B_op_dict = other._operators_dict

        # αβ + β ∑ᵢAᵢ
        self.__imul__(β)

        # α ∑ᵢBᵢ
        if np.abs(α) > self.mel_cutoff:
            for aon, op in B_op_dict.items():
                self._add_operator(aon, α * op)

        # ∑ᵢⱼAᵢBⱼ
        for supp_A_i, A_i in A_op_dict.items():
            for supp_B_j, B_j in B_op_dict.items():
                self._add_operator(
                    *_multiply_operators(
                        self.hilbert, supp_A_i, A_i, supp_B_j, B_j, dtype=self.dtype
                    )
                )

        is_hermitian_A = self._is_hermitian
        is_hermitian_B = other._is_hermitian

        self._reset_caches()

        if is_hermitian_A and is_hermitian_B:
            self._is_hermitian = is_hermitian_A

        return self

    def _reset_caches(self):
        """
        Cleans the internal caches built on the operator.
        """
        self._initialized = False
        self._is_hermitian = None

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        self._setup()
        return self._max_conn_size

    def __repr__(self):
        ao = self.acting_on

        acting_str = f"acting_on={ao}"
        if len(acting_str) > 55:
            acting_str = f"#acting_on={len(ao)} locations"
        return f"{type(self).__name__}(dim={self.hilbert.size}, {acting_str}, constant={self.constant}, dtype={self.dtype})"



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import numpy as np
from flax.serialization import to_bytes
from flax.core import pop as fpop, FrozenDict

from netket.logging import AbstractLog
from netket.jax.sharding import extract_replicated

_mode_shorthands = {"write": "w", "append": "a", "fail": "x"}


def tree_log(tree, root, data, *, iter=None):
    """
    Maps all elements in tree, recursively calling tree_log with a new root string,
    and when it reaches leaves adds them to the `data` inplace.

    Args:
        tree: a pytree where the leaf nodes contain data
        root: the root of the tags used to log to HDF5
        data: an HDF5 file modified in place
        iter: an integer number specifying at which iteration the data was generated
    """

    if tree is None:
        return

    elif isinstance(tree, list):
        for i, val in enumerate(tree):
            tree_log(val, f"{root}/{i}", data, iter=iter)

    # handle namedtuples
    elif isinstance(tree, list) and hasattr(tree, "_fields"):
        tree_log(iter, f"{root}/iter", data)
        for key in tree._fields:
            tree_log(getattr(tree, key), f"{root}/{key}", data)

    elif isinstance(tree, tuple):
        tree_log(iter, f"{root}/iter", data)
        for i, val in enumerate(tree):
            tree_log(val, f"{root}/{i}", data)

    elif isinstance(tree, dict):
        for key, value in tree.items():
            tree_log(value, f"{root}/{key}", data, iter=iter)  # noqa: F722

    elif hasattr(tree, "to_compound"):
        tree_log(iter, f"{root}/iter", data)
        tree_log(tree.to_compound()[1], root, data)  # noqa: F722

    elif hasattr(tree, "to_dict"):
        tree_log(iter, f"{root}/iter", data)
        tree_log(tree.to_dict(), root, data)  # noqa: F722

    else:
        if iter is not None:
            tree_log(iter, f"{root}/iter", data)
            root = f"{root}/value"
        value = np.asarray(tree)
        if root in data:
            f_value = data[root]
            f_value.resize(f_value.shape[0] + 1, axis=0)
            f_value[-1] = value
        else:
            maxshape = (None, *value.shape)
            data.create_dataset(root, data=[value], maxshape=maxshape)


class HDF5Log(AbstractLog):
    r"""
    HDF5 Logger, that can be passed with keyword argument `logger` to Monte
    Carlo drivers in order to serialize the output data of the simulation.

    The logger has support for scalar numbers, NumPy/JAX arrays, and netket.stats.Stats objects.
    These are stored as individual groups within a HDF5 file, under the main group `data/`:

    - scalars are stored as a group with one dataset values of shape :code:`(n_steps,)` containing the logged values,
    - arrays are stored in the same way, but with values having shape :code:`(n_steps, *array_shape)`,
    - netket.stats.Stats are stored as a group containing each field :code:`(Mean, Variance, etc...)` as a separate dataset.

    Importantly, each group has a dataset :code:`iters`, which tracks the
    iteration number of the logged quantity.

    If the model state is serialized, then it is serialized as a dataset in the group `variational_state/`.
    The target of the serialization is the parameters PyTree of the variational state (stored in the group
    `variational_state/parameters`), and the rest of the variational state variables (stored in the group
    `variational_state/model_state`)

    Data can be deserialized by calling :code:`f = h5py.File(filename, 'r')` and
    inspecting the datasets as a dictionary, i.e. :code:`f['data/energy/Mean']`

    .. note::
        The API of this logger is covered by our Semantic Versioning API guarantees. However, the structure of the
        logged files is not, and might change in the future while we iterate on this logger. If you think that we
        could improve the output format of this logger, please open an issue on the NetKet repository and let us
        know.

    """

    def __init__(
        self,
        path: str,
        mode: str = "write",
        save_params: bool = True,
        save_params_every: int = 1,
    ):
        """
        Construct a HDF5 Logger.

        Args:
            path: the name of the output files before the extension
            mode: Specify the behaviour in case the file already exists at this
                path. Options are
                - `[w]rite`: (default) overwrites file if it already exists;
                - `[x]` or `fail`: fails if file already exists;
            save_params: bool flag indicating whether variables of the variational state
                should be serialized at some interval
            save_params_every: every how many iterations should machine parameters be
                flushed to file
        """
        import h5py  # noqa: F401

        super().__init__()

        if not ((mode == "write") or (mode == "append") or (mode == "fail")):
            raise ValueError(
                "Mode not recognized: should be one of `[w]rite`, `[a]ppend` or"
                "`[x]`(fail)."
            )
        mode = _mode_shorthands[mode]

        if not path.endswith((".h5", ".hdf5")):
            path = path + ".h5"

        if os.path.exists(path) and mode == "x":
            raise ValueError(
                "Output file already exists. Either delete it manually or"
                "change `path`."
            )

        dir_name = os.path.dirname(path)
        if dir_name != "":
            os.makedirs(dir_name, exist_ok=True)

        self._file_mode = mode
        self._file_name = path
        self._writer = None

        self._save_params = save_params
        self._save_params_every = save_params_every
        self._steps_notsaved_params = 0

    def _init_output_file(self):
        if self._is_master_process:
            import h5py

            self._writer = h5py.File(self._file_name, self._file_mode)

    def __call__(self, step, log_data, variational_state):
        if self._writer is None:
            self._init_output_file()

        if self._steps_notsaved_params % self._save_params_every == 0:
            variables = variational_state.variables

            # TODO: remove - FrozenDict are deprecated
            if isinstance(variables, FrozenDict):
                variables = variables.unfreeze()
            variables = extract_replicated(variables)
            self._steps_notsaved_params = 0
        else:
            variables = None

        if self._is_master_process:
            tree_log(log_data, "data", self._writer, iter=step)

            if variables is not None:
                _, params = fpop(variables, "params")
                binary_data = to_bytes(variables)
                tree = {"model_state": binary_data, "parameters": params, "iter": step}
                tree_log(tree, "variational_state", self._writer)
            self._writer.flush()

        self._steps_notsaved_params += 1

    def flush(self, variational_state=None):
        """
        Writes to file the content of this logger.

        Args:
            variational_state: optionally also writes the parameters of the machine.
        """
        if self._writer is not None:
            self._writer.flush()

    def __del__(self):
        if hasattr(self, "_writer"):
            self.flush()

    def __repr__(self):
        _str = f"HDF5Log('{self._file_name}', mode={self._file_mode}"
        return _str



# utilities to prepare the internal datastructures


from typing import Any, Union

from functools import partial
import numpy as np

import jax
import jax.numpy as jnp

from netket.jax import COOArray
from netket.utils.types import Array

import sparse

from .._normal_order_utils import SpinOperatorArrayTerms

PNCOperatorDataType = tuple[Union[Array, COOArray, None], Union[Array, None], Array]
r"""
Custom sparse internal data for ParticleNumberConservingFermioperator2nd of strings of fermionic operators of a fixed length

It is given by a 3-tuple for a fixed length N of string in normal ordering (containg 2N fermionic operators),

Assume we are given a sequence of equal-length normal ordered strings :math:`\sum_i w_i \hat c_{a_{i1}}^\dagger \cdots \hat c_{a_{iN}}^\dagger \hat c_{b_{i1}} \cdots \hat c_{b_{iN}}`
with :math:`2N` fermionic operators, with larger indices to the left: :math:`a_{i1} >=\cdots>=a_{iN}, b_{i1} >=\cdots>=b_{iN}`

Here :math:`a \in \{1 \dots n\}`` contains the indices of the :math:`\hat c^\dagger`, b the indices of the c operators and w the corresponding weight of every string.
where n is the number of orbitals.

The 3-tuple for these operators is given by
    index_array: shape (n,)*N+(n_max,) contains list of integer indices for all the strings for a given list of destruction operators (given by b above)
                 can be stored either in a dense, or sparse format; is padded to the maximum number of operators n_max for a given sequence of destruction ops (b)
                 Since the operators are assumed to be in normal order (larger sites to the left) only the lower triangular part is used.
    create_array: shape (n_ops+1,)+(N,) for every index from index_array contains creation operators of the corresponding term (given by a above)
    weight_array: shape (n_ops+1,)+(1,) for every index from index_array contains the weight of the corresponding term
                  The 0 weight for the padding is stored as the last element.

We optionally treat the diagonal opeators separately to the non-diagonal ones in a more efficient way:
    index_array: None
    create_array: None
    weight_array: shape (n,)*N  is indexed directly with the list of destruction operators
"""

PNCOperatorDataDict = dict[Union[int, tuple[int, tuple[int]]], PNCOperatorDataType]
r"""
Custom sparse internal data for ParticleNumberConservingFermioperator2nd of strings of fermionic operators of different lengths N
"""

PNCOperatorDataCollectionDict = dict[str, PNCOperatorDataDict]
r"""
collection of different PNCOperatorDataDict, each with a name. E.g. for storing the diagonal and off-diagonal part of an operator separately
"""


PNCOperatorArrayTerms = tuple[Array, Array]
r"""
particle-number conserving equivalent of OperatorArrayTerms, without daggers array
First Half of the operators are assumed to be creation, and the second half destruction
A tuple (sites, weights) where
    sites: An integer array of size n_terms x n_operators containing the indices
    weights: An array of size n_terms containing the weight of each term
"""

PNCOperatorArrayDict = dict[int, PNCOperatorArrayTerms]
r"""
A dictionary containing OperatorArrayTerms of different lengths, where the key specifies the lenght,
representing a particle-number conserving fermionic operator in second quantization.
"""


CoordsDataDictSectorType = dict[tuple[int, tuple[int]], tuple[Array, Array]]
r"""
a dict {(k, sectors) : (indices, data)}
where k is the number of c/c^\dagger, sectors are the spin sectors acted on,
indices contains the sites inside of each sector and data contains the weights
"""

CoordsDataDictSectorCooArrayType = dict[tuple[int, tuple[int]], sparse.COO]
r"""
a dict {(k, sectors) : array }
where k is the number of c/c^\dagger, sectors are the spin sectors acted on,
and array (of size n_orbitals^k) contains the indices and weights of the operators
"""


def _len_helper(s: Any) -> int:
    """
    return the length of the argument if it is a tuple, else return 1
    """
    if isinstance(s, tuple):
        return len(s)
    else:
        return 1


def _prepare_data_helper(
    sites_destr: Array,
    sites_create: Array,
    weights: Array,
    n_orbitals: int,
    sparse_: bool = True,
) -> PNCOperatorDataType:
    r"""
    Helper function to prepare the sparse operator

    Args:
        sites_destr:  indices of the :math:`\hat c`
        sites_create: indices of the :math:`\hat c^\dagger`
        weights: weights of each term
        n_orbitals: number of orbitals
        sparse_: use COOArray for the index array
    Returns:
        sparse operator data
    """

    # we encode sites_create==sites_destr by passing sites_create=None
    is_diagonal = sites_create is None

    n_terms, half_n_ops = sites_destr.shape
    assert weights.shape == (n_terms,)
    if not is_diagonal:
        assert sites_destr.shape == sites_create.shape

    if half_n_ops == 0:  # constant
        assert n_terms == 1
        index_array = jnp.zeros((), dtype=np.int32)
        create_array = jnp.zeros((1, 1, 0), dtype=np.int32)
        weight_array = jnp.array(weights, dtype=weights.dtype)
    elif is_diagonal:
        assert sites_destr.max() < n_orbitals
        index_array = None
        create_array = None
        # use sparse.COO to sort since COOArray expects sorted indices
        # TODO do it inside COOArray
        tmp = sparse.COO(sites_destr.T, weights, (n_orbitals,) * (half_n_ops))
        weight_array = COOArray(
            jnp.asarray(tmp.coords.T),
            jnp.asarray(tmp.data),
            (n_orbitals,) * (half_n_ops),
        )
        if not sparse_:
            weight_array = weight_array.todense()
    else:
        assert sites_destr.max() < n_orbitals
        assert sites_create.max() < n_orbitals

        ### simple, inefficient version
        # destr_unique = np.unique(sites_destr, axis=0)
        # nper = np.zeros(len(destr_unique), dtype=int)
        # for i, d in enumerate(destr_unique):
        #     nper[i] = (d[None] == sites_destr).all(axis=-1).sum()
        ###
        A = sparse.COO(
            np.concatenate([sites_destr, sites_create], axis=1).T,
            weights,
            shape=(n_orbitals,) * (2 * half_n_ops),
        )
        axes_create = tuple(range(A.ndim // 2, A.ndim))
        n_destr = (A != 0).sum(axes_create)
        destr_unique = n_destr.coords.T
        nper = n_destr.data
        ###

        # we pad with zeros, so we take create_array and weight_array of size nunique+1
        # (where the 0th element is the padding)
        # and put zeros in the index_array, for terms which dont exist

        ### simple, inefficient version
        # nmax = int(nper.max())
        # nunique = len(destr_unique)
        # create_array = jnp.zeros((1 + nunique, nmax, half_n_ops), dtype=np.int32)
        # weight_array = jnp.zeros((1 + nunique, nmax), dtype=weights.dtype)
        # for i, d in enumerate(destr_unique):
        #     mask = (sites_destr == d).all(axis=-1)
        #     weight_array = weight_array.at[i + 1, : nper[i]].set(weights[mask])
        #     create_array = create_array.at[i + 1, : nper[i]].set(sites_create[mask])
        ###
        # select only nonzero destr rows
        B = A[tuple(destr_unique.T)]
        # create an arange for every row
        row_ind, *create_indices = B.coords
        row_start = np.where(np.diff(row_ind, prepend=-1))
        a = np.arange(B.nnz)
        compressed_col_ind = a - np.repeat(a[row_start], nper)
        # +1 because of the padding
        new_coords = np.vstack([row_ind + 1, compressed_col_ind])
        weight_array = jnp.asarray(
            sparse.COO(new_coords, B.data, shape=new_coords.max(axis=1) + 1).todense()
        )
        create_array = jnp.concatenate(
            [
                sparse.COO(new_coords, c, shape=weight_array.shape).todense()[..., None]
                for c in create_indices
            ],
            axis=-1,
        )
        ###
        # destr_unique should be already sorted at this point (in np.unique / sparse.COO)
        index_array = COOArray(
            jnp.asarray(destr_unique),
            jnp.arange(1, len(destr_unique) + 1),
            (n_orbitals,) * (half_n_ops),
        )
        if not sparse_:
            index_array = index_array.todense()

    return index_array, create_array, weight_array


def prepare_data_diagonal(
    sites_destr: Array, weights: Array, n_orbitals: int, **kwargs
) -> PNCOperatorDataType:
    r"""

    Prepare the custom sparse internal data for ParticleNumberConservingFermioperator2nd, for the diagonal part of the operator
    of strings of a fixed length

    Assume we are given a sequence of equal-length normal ordered strings \sum_i w_i c_{a_i1}^\dagger ... c_{a_iN}^\dagger c_{a_i1} ... c_{a_iN}
    with 2N fermionic operators, with larger indices to the left: a_i1 >=...>=a_iN

    Please refer to the docstring of prepare_data for a more complete explanation of the storage format.

    We can treat the diagonal opeators separately to the non-diagonal ones in a more efficient way:
        index_array: None
        create_array: None
        weight_array: shape (n,)*N  is indexed directly with the list of destruction operators (given by b above)

    Args:
        sites_destr: a matrix containing the indices of c^dagger/c for every string [[a_i1, ..., a_iN]] ()
        weights: array of the corresponding weights [w_i]
        n_orbitals: number of orbitals n
        sparse_: whether to store weight_array in dense or sparse
    Returns:
        A tuple (index_array, create_array, weight_array) as defined above

    """
    return _prepare_data_helper(sites_destr, None, weights, n_orbitals, **kwargs)


def prepare_data(
    sites: Array, weights: Array, n_orbitals: int, **kwargs
) -> PNCOperatorDataType:
    r"""
    Prepare the custom sparse internal data for ParticleNumberConservingFermioperator2nd
    of strings of a fixed length

    It is given by a 3-tuple for every length N of string in normal ordering (containg 2N fermionic operators),

    Assume we are given a sequence of equal-length normal ordered strings \sum_i w_i c_{a_i1}^\dagger ... c_{a_iN}^\dagger c_{b_i1} ... c_{b_iN}
    with 2N fermionic operators, with larger indices to the left: a_i1 >=...>=a_iN, b_i1 >=...>=b_iN

    Here a \in {1..n} contains the indices of the c^\dagger, b the indices of the c operators and w the corresponding weight of every string.
    where n is the number of orbitals.

    The 3-tuple for these operators is given by
        index_array: shape (n,)*N+(n_max,) contains list of integer indices for all the strings for a given list of destruction operators (given by b above)
                     can be stored either in a dense, or sparse format; is padded to the maximum number of operators n_max for a given sequence of destruction ops (b)
                     Since the operators are assumed to be in normal order (larger sites to the left) only the lower triangular part is used.
        create_array: shape (n_ops+1,)+(N,) for every index from index_array contains creation operators of the corresponding term (given by a above)
        weight_array: shape (n_ops+1,)+(1,) for every index from index_array contains the weight of the corresponding term
                      The 0 weight for the padding is stored as the last element.

    Then for a given basis state |b_1,...,b_m> (where b_j indicates the occupied orbitals, for a fixed number of electrons m) we find all the connected elements
    by taking all m choose N combinations of occupied orbitals to be destroyed as index for index_array, excluding the strings which try to destroy an empty orbital.


    Args:
        sites: a matrix containing the indices of c^dagger and c for every string [[a_i1, ..., a_iN, b_i1, ..., b_iN]]
        weights: array of the corresponding weights [w_i]
        n_orbitals: number of orbitals n
        sparse_: wether to store index_array in dense or sparse
    Returns:
        A tuple (index_array, create_array, weight_array) as defined above

    """

    # sites is an array (n_terms, n_ops) containing the sites
    # of terms in normal order (daggers to left, desc order)
    #
    # weights is of shape (n_terms,)
    n_terms, n_ops = sites.shape
    assert n_ops % 2 == 0
    sites_destr = sites[:, : n_ops // 2]
    sites_create = sites[:, n_ops // 2 :]
    return _prepare_data_helper(
        sites_destr, sites_create, weights, n_orbitals, **kwargs
    )


def split_diag_offdiag(
    sites: Array, weights: Array
) -> tuple[tuple[Array, Array], tuple[Array, Array]]:
    r"""
    Split the diagonal and off-diagonal terms

    Args:
        sites: indices of the :math:`\hat c^\dagger` and :math:`\hat c`, stored in the first and second half of the last axis of the array
        weights: corresponding weights
    Returns:
        sites and weights of the diagonal and off-diagonal terms
    """
    n_terms, n_ops = sites.shape
    assert weights.shape == (n_terms,)
    assert n_ops % 2 == 0

    idestr = sites[:, : (n_ops // 2)]
    icreate = sites[:, (n_ops // 2) :]

    is_diag = (idestr == icreate).all(axis=-1)
    diag_sites = idestr[is_diag]
    diag_weights = weights[is_diag]
    offdiag_sites = sites[~is_diag]
    offdiag_weights = weights[~is_diag]
    return (diag_sites, diag_weights), (offdiag_sites, offdiag_weights)


def prepare_operator_data_from_coords_data_dict(
    coords_data_dict: PNCOperatorArrayDict, n_orbitals: int, **kwargs
) -> PNCOperatorDataCollectionDict:
    r"""
    Prepare the custom sparse internal data for ParticleNumberConservingFermioperator2nd

    of a string of operators \sum_N \sum_i w_i^{(N)} c_{a_i1^{(N)}}^\dagger ... c_{a_iN^{(N)}}^\dagger c_{b_i1^{(N)}} ... c_{b_iN^{(N)}}
    in descending order a_i1 >=...>=a_iN, b_i1 >=...>=b_iN.

    Please refer to the docstring of prepare_data, prepare_data_diagonal for a more complete explanation of the storage format.

    Args:
        coords_data_dict: A dictionary {N: (sites, weights)}
            where for every length N
                sites is a matrix containing the stacked indices [[a_i1^{(N)}, ... a_iN^{(N)}, b_i1^{(N)}, ..., b_iN^{(N)}]]
                of the c^\dagger and c
                weights is a vector containing the corresponding weights [w_i^{(N)}]
        n_orbitals: number of orbitals
    Returns:
        A dictionary {'diag': { N:  (None, None, weight_array)}, 'offdiag': { N : (index_array, create_array, weight_array)}}
        containing the sparse representation for every lenght N of strings c_{a_i1}^\dagger ... c_{a_iN}^\dagger c_{b_i1} ... c_{b_iN}
    """
    data_offdiag = {}
    data_diag = {}
    for k, v in coords_data_dict.items():
        sw_diag, sw_offdiag = split_diag_offdiag(*v)
        if len(sw_diag[-1]) > 0:
            data_diag[k] = prepare_data_diagonal(*sw_diag, n_orbitals, **kwargs)
        if len(sw_offdiag[-1]) > 0:
            data_offdiag[k] = prepare_data(*sw_offdiag, n_orbitals, **kwargs)
    data = {"diag": data_diag, "offdiag": data_offdiag}
    return data


def sparse_arrays_to_coords_data_dict(
    ops: dict[Any, sparse.COO],
) -> dict[Any, tuple[Array, Array]]:
    r"""
    Split dictionary values of sparse arrays into indices and data

    (optionally) supports spin sectors in the key
    """

    def _unpack(k, v):
        if isinstance(k, tuple):
            k = k[0]  # rest are e.g. the spin sectors
        if k == 0:
            return np.zeros((1, 0), dtype=int), np.array([v])
        else:
            return v.coords.T, v.data

    return {k: _unpack(k, v) for k, v in ops.items()}


def collect_ops(
    operators: list[Union[Array, sparse.COO]],
) -> dict[int, Union[Array, sparse.COO]]:
    r"""
    sum operators with the same number of c/c^\dagger

    Args:
        operators: a list of scalars / sparse matrices / dense matrices
    Returns:
        A dictionary of sparse matrices, one for each lenght of c/c^\dagger
    """

    ops = {}
    for A in operators:
        if isinstance(A, sparse.COO):
            k = A.ndim
            if A.shape == ():
                A = A.fill_value
            else:
                assert A.fill_value == 0
        elif jnp.isscalar(A) or (hasattr(A, "__array__") and A.ndim == 0):
            A = np.asarray(A)
            k = 0
        elif hasattr(A, "__array__"):
            A = sparse.COO.from_numpy(np.asarray(A))
            k = A.ndim
        else:
            raise NotImplementedError
        Ak = ops.pop(k, None)
        if Ak is not None:
            ops[k] = Ak + A
        else:
            ops[k] = A
    return ops


def prepare_operator_data_from_coords_data_dict_spin(
    coords_data_sectors: CoordsDataDictSectorType, n_orbitals: int
) -> PNCOperatorDataCollectionDict:
    r"""
    Prepare the custom sparse internal data for ParticleNumberConservingFermioperator2nd
    of a string of operators :math:`\sum_N \sum_i \sum_s  w_i^{(N)} \hat c_{a_{i1}^{(N)},s}^\dagger \cdots \hat c_{a_{iN}^{(N)},s}^\dagger \hat c_{b_{i1}^{(N)},s} \cdots \hat c_{b_{iN}^{(N)},s}`
    in descending order :math:`a_{i1} >=...>=a_{iN}, b_{i1} >=...>=b_{iN}`.

    Version with spin sectors.

    Args:
        coords_data_sectors: A dictionary {(N, sectors): (sites, weights)}
            where for every length N, and tuple of spin sectors
                sites is a matrix containing the stacked indices [[a_i1^{(N)}, ... a_iN^{(N)}, b_i1^{(N)}, ..., b_iN^{(N)}]]
                of the c^\dagger and c
                weights is a vector containing the corresponding weights [w_i^{(N)}]
        n_orbitals: number of orbitals
    Returns:
        A dictionary {'diag': { N:  (None, None, weight_array)}, 'offdiag': { N : (index_array, create_array, weight_array)}}
        containing the sparse representation for every lenght N of strings c_{a_i1}^\dagger ... c_{a_iN}^\dagger c_{b_i1} ... c_{b_iN}
    """

    # version with sectors
    _cond = lambda s: s == () or _len_helper(s[0]) == 1
    coords_data_same = {
        (k, s): v for (k, s), v in coords_data_sectors.items() if _cond(s)
    }
    coords_data_mixed = {
        (k, s): v for (k, s), v in coords_data_sectors.items() if not _cond(s)
    }

    operator_data = prepare_operator_data_from_coords_data_dict(
        coords_data_same, n_orbitals
    )
    # process mixed terms
    data_diag_mixed = {}
    data_offdiag_mixed = {}
    for k, v in coords_data_mixed.items():
        sw_diag, sw_offdiag = split_diag_offdiag(*v)
        if len(sw_diag[-1]) > 0:
            data_diag_mixed[k] = prepare_data_diagonal(
                *sw_diag, n_orbitals, sparse_=False
            )
        if len(sw_offdiag[-1]) > 0:
            data_offdiag_mixed[k] = prepare_data(*sw_offdiag, n_orbitals, sparse_=False)
    operator_data = {
        **operator_data,
        "mixed_diag": data_diag_mixed,
        "mixed_offdiag": data_offdiag_mixed,
    }
    return operator_data


def sites_daggers_weights_to_sparse(
    sites: Array, daggers: Array, weights: Array, n_orbitals: int
) -> sparse.COO:
    r"""
    Convert normal ordered
    Args:
        sites:
        daggers: first half of the last axis needs to be 1 (create), second half 0 (destroy)
        weights:
    Returns:
        a sparse.COO tensor

    """
    n = daggers.shape[-1]
    assert n % 2 == 0
    assert (daggers[:, : n // 2] == 1).all()
    assert (daggers[:, n // 2 :] == 0).all()
    return sparse.COO(sites.T, weights, shape=(n_orbitals,) * n)


def _insert_append_helper(
    d: CoordsDataDictSectorCooArrayType,
    k: tuple[int, tuple[int]],
    s: tuple[int],
    o: Union[sparse.COO, Array],
    cutoff: float,
):
    r"""
    check if an element with the same matrix but different sectors exist
    if yes append to the list of sectors, else insert new element into the dict
    """
    for (k2, s2), o2 in d.items():
        same_number_of_sectors = (s == () and s2 == ()) or (
            len(s2) > 0 and len(s) > 0 and _len_helper(s2[0]) == _len_helper(s[0])
        )
        if (
            same_number_of_sectors
            and k == k2  # same_number_of_fermionic_operators
            and sparse.abs(o - o2).max() < cutoff  # same_matrix
        ):
            d[k, s2 + s] = d.pop((k2, s2))
            break
    else:
        d[k, s] = o


def to_coords_data_sector(
    tno_sector: SpinOperatorArrayTerms,
    n_spin_subsectors: int,
    n_orbitals: int,
    cutoff: float = 1e-11,
) -> CoordsDataDictSectorType:
    r"""
    Args:
        tno_sector: a list of tuples [(sites, sectors, daggers, weights)]
                    of terms in normal order with higher sectors on the left

    Returns: a dict {(k, sectors) : (indices, data)}
             where k is the number of c/c^\dagger,
             sectors are the spin sectors acted on,
             indices contains the sites inside of each sector
             and data contains the weights
    """

    operators_sector = {}

    for k, (sites, sectors, daggers, weights) in tno_sector.items():
        for i in range(n_spin_subsectors):
            if not (((2 * daggers - 1) * (sectors == i)).sum(axis=-1) == 0).all():
                raise ValueError  # does not conserve particle number per sector

        sector_count = jax.vmap(partial(jnp.bincount, length=n_spin_subsectors))(
            sectors
        )

        # merge sectors which have same sparse matrix

        if k == 0:
            operators_sector[0, ()] = weights.reshape(())
        elif k == 2:
            # at this point we know there is only one sector this acts on
            sector = sectors[:, 0]  # = sectors[:, 1]
            for i in np.unique(sector):
                m = sector == i
                o = sites_daggers_weights_to_sparse(
                    sites[m], daggers[m], weights[m], n_orbitals=n_orbitals
                )
                _insert_append_helper(operators_sector, k, (i,), o, cutoff)
        elif k == 4:
            # at this point we know that n_sectors_acting_on \in 1,2
            n_sectors_acting_on = np.count_nonzero(sector_count, axis=-1)

            # all same sector
            m_same = n_sectors_acting_on == 1
            sector = sectors[:, 0]
            for i in np.unique(sector[m_same]):
                m = (sector == i) & m_same
                o = sites_daggers_weights_to_sparse(
                    sites[m], daggers[m], weights[m], n_orbitals=n_orbitals
                )
                _insert_append_helper(operators_sector, k, (i,), o, cutoff)

            m_different = ~m_same
            sector = sectors[:, :2]
            # i > j because we made it normal order (with site shifted by N*spin) above
            for ij in np.unique(sector[m_different], axis=0):
                m = (sector == ij[None]).all(axis=-1) & m_different
                # minus sign because in the operator (_get_conn_padded_interaction_up_down) we assume it's swaped to (assuming σ>ρ)
                # cσ^† cσ cρ^† cρ = - cσ^† cρ^† cσ cρ
                o = -sites_daggers_weights_to_sparse(
                    sites[m], daggers[m], weights[m], n_orbitals=n_orbitals
                )
                _insert_append_helper(operators_sector, k, (tuple(ij),), o, cutoff)
        else:
            raise NotImplementedError

    return sparse_arrays_to_coords_data_dict(operators_sector)
